import uuid
from contextlib import contextmanager
from logging import Logger
from typing import Iterator, Optional
from uuid import UUID

from layerapi.api.entity.project_pb2 import Project as ProjectMessage
from layerapi.api.entity.project_view_pb2 import ProjectView
from layerapi.api.ids_pb2 import ProjectId
from layerapi.api.service.flowmanager.project_api_pb2 import (
    CreateProjectRequest,
    GetProjectByPathRequest,
    GetProjectByPathResponse,
    GetProjectViewByIdRequest,
    GetProjectViewByIdResponse,
    RemoveProjectByIdRequest,
    UpdateProjectRequest,
)
from layerapi.api.service.flowmanager.project_api_pb2_grpc import ProjectAPIStub

from layer.config import ClientConfig
from layer.contracts.accounts import Account
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.projects import Project
from layer.exceptions.exceptions import (
    LayerClientResourceAlreadyExistsException,
    LayerClientResourceNotFoundException,
)
from layer.utils.grpc import create_grpc_channel, generate_client_error_from_grpc_error


class ProjectServiceClient:
    _service: ProjectAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
    ):
        self._config = config.project_service
        self._logger = logger
        self._access_token = config.access_token
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

    @contextmanager
    def init(self) -> Iterator["ProjectServiceClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = ProjectAPIStub(channel=channel)
            yield self

    @staticmethod
    def _map_project_message_to_project_contract(
        full_name: ProjectFullName, project_msg: ProjectMessage
    ) -> Project:
        project_id = UUID(project_msg.id.value)
        account_id = UUID(project_msg.organization_id.value)
        return Project(
            name=full_name.project_name,
            id=project_id,
            account=Account(
                name=full_name.account_name,
                id=account_id,
            ),
        )

    @staticmethod
    def _map_project_view_message_to_project_contract(
        project_view: ProjectView,
    ) -> Project:
        project_id = UUID(project_view.id.value)
        account_id = UUID(project_view.account.id.value)
        return Project(
            name=project_view.name,
            id=project_id,
            account=Account(
                name=project_view.account.name,
                id=account_id,
            ),
        )

    def get_project_by_id(self, project_id: UUID) -> Optional[Project]:
        try:
            resp: GetProjectViewByIdResponse = self._service.GetProjectViewById(
                GetProjectViewByIdRequest(project_id=ProjectId(value=str(project_id)))
            )
            if resp.project is not None:
                return self._map_project_view_message_to_project_contract(resp.project)
        except LayerClientResourceNotFoundException:
            pass
        except Exception as err:
            raise generate_client_error_from_grpc_error(err, "internal")
        return None

    def get_project(self, full_name: ProjectFullName) -> Optional[Project]:
        try:
            resp: GetProjectByPathResponse = self._service.GetProjectByPath(
                GetProjectByPathRequest(path=full_name.path)
            )
            if resp.project is not None:
                return self._map_project_message_to_project_contract(
                    full_name, resp.project
                )
        except LayerClientResourceNotFoundException:
            pass
        except Exception as err:
            raise generate_client_error_from_grpc_error(err, "internal")
        return None

    def remove_project(self, project_id: uuid.UUID) -> None:
        self._service.RemoveProjectById(
            RemoveProjectByIdRequest(project_id=ProjectId(value=str(project_id)))
        )

    def create_project(self, full_name: ProjectFullName) -> Project:
        try:
            resp = self._service.CreateProject(
                CreateProjectRequest(
                    project_full_name=full_name.path,
                    visibility=ProjectMessage.VISIBILITY_PRIVATE,
                )
            )
            return self._map_project_message_to_project_contract(
                full_name, resp.project
            )
        except LayerClientResourceAlreadyExistsException as e:
            raise e
        except Exception as err:
            raise generate_client_error_from_grpc_error(err, "internal")

    def update_project_readme(
        self, project_full_name: ProjectFullName, readme: str
    ) -> None:
        try:
            self._service.UpdateProject(
                UpdateProjectRequest(
                    project_full_name=project_full_name.path, readme=readme
                )
            )
        except LayerClientResourceNotFoundException as e:
            raise e
        except Exception as err:
            raise generate_client_error_from_grpc_error(err, "internal")

    def set_project_visibility(
        self, project_full_name: ProjectFullName, *, is_public: bool
    ) -> None:
        visibility = (
            ProjectMessage.VISIBILITY_PUBLIC
            if is_public
            else ProjectMessage.VISIBILITY_PRIVATE
        )
        try:
            self._service.UpdateProject(
                UpdateProjectRequest(
                    project_full_name=project_full_name.path, visibility=visibility
                )
            )
        except LayerClientResourceNotFoundException as e:
            raise e
        except Exception as err:
            raise generate_client_error_from_grpc_error(err, "internal")



from pathlib import Path
from typing import Optional

from layer.config import DEFAULT_LAYER_PATH


class CacheError(Exception):
    pass


class Cache:
    def __init__(self, cache_dir: Optional[Path] = None) -> None:
        self._cache_dir = (cache_dir or DEFAULT_LAYER_PATH) / "cache"

    @property
    def cache_dir(self) -> Path:
        return self._cache_dir

    @property
    def is_initialised(self) -> bool:
        return self._cache_dir.exists()

    def _ensure_initialised(self) -> None:
        if not self.is_initialised:
            raise CacheError("Cache not initalised")

    def initialise(self) -> "Cache":
        self._cache_dir.mkdir(parents=True, exist_ok=True)
        return self

    def put_path_entry(self, key: str, path: Path) -> Optional[Path]:
        self._ensure_initialised()
        if not path.exists():
            return None
        abs_path = path.absolute()
        key_path = abs_path.parent / self._cache_dir / key
        abs_path.rename(key_path)
        return key_path

    def get_path_entry(self, key: str) -> Optional[Path]:
        self._ensure_initialised()
        if not key:
            return None
        key_path = self._cache_dir / key
        if key_path.exists():
            return key_path
        return None

    def clear(self) -> None:
        if self._cache_dir.exists():
            import shutil

            shutil.rmtree(self._cache_dir.absolute().as_posix())



from abc import ABCMeta
from typing import Any, Callable, Dict, List

import wrapt  # type: ignore

from layer.decorators.layer_wrapper import LayerFunctionWrapper


def fabric(name: str) -> Callable[..., Any]:
    """
    Allows specifying the resources the Layer backend should use to run decorated function.

    Note: The fabric decorator should only ever be used alongside Layer `dataset` or `model` decorators. GPU fabrics are available for model training only and can only be used in combination with `model` decorator.

    This decorator overrides the fabric set via `layer.init`. It has no impact on local execution of your functions, only on remote execution in the Layer backend.

    :param name: `Name of the fabric </docs/reference/fabrics#predefined-fabrics>`_ to use in Layer backend.
    :return: Function object.

    .. code-block:: python

        import layer
        from layer.decorators import dataset, fabric

        # fabric below will determine resources used to run `create_my_dataset` function in Layer backend
        @fabric("f-medium")
        @dataset("product-data")
        def create_product_dataset():
            data = [[1, "product1", 15], [2, "product2", 20], [3, "product3", 10]]
            dataframe = pd.DataFrame(data, columns=["Id", "Product", "Price"])
            return dataframe

        # fabric setting will be used when running your function below
        layer.run(create_product_dataset)
    """

    @wrapt.decorator(proxy=_fabric_wrapper(name))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> Any:
        return wrapped(*args, **kwargs)

    return wrapper


def _fabric_wrapper(fabric_name: str) -> Any:
    class FabricFunctionWrapper(LayerFunctionWrapper, metaclass=ABCMeta):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any = None) -> None:
            super().__init__(wrapped, wrapper, enabled)
            self.layer.set_fabric(fabric_name)

    return FabricFunctionWrapper



import urllib.request
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import pandas
import pyarrow
from layerapi.api.service.dataset.dataset_api_pb2 import Command
from layerapi.api.value.ticket_pb2 import PartitionTicket
from pyarrow import flight as fl
from pyarrow.lib import ArrowKeyError

from layer.cache.cache import Cache
from layer.utils.grpc import create_grpc_ssl_config


class DatasetClientError(Exception):
    pass


def _dataset_exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:
    def inner_function(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except ArrowKeyError as e:  # pytype: disable=mro-error
            not_found_message = str(e).replace(
                "gRPC returned not found error, with message: ",
                "",
            )
            raise DatasetClientError(not_found_message)
        except DatasetClientError as e:
            raise e
        except Exception as e:
            raise DatasetClientError("Query failed") from e

    return inner_function


@dataclass(frozen=True)
class PartitionMetadata:
    location: str
    format: int = PartitionTicket.Format.FORMAT_PARQUET
    checksum: str = ""


class Partition:
    def __init__(
        self,
        reader: Union[fl.FlightStreamReader, pandas.DataFrame],
        from_cache: bool = False,
    ):
        self._reader = reader
        self._from_cache = from_cache

    def to_pandas(self) -> pandas.DataFrame:
        if isinstance(self._reader, pandas.DataFrame):
            return self._reader
        return self._reader.read_pandas()

    @property
    def from_cache(self) -> bool:
        return self._from_cache


class _FlightCallMetadataMiddleware(fl.ClientMiddleware):
    def __init__(self, call_metadata: List[Tuple[str, str]], client: "DatasetClient"):
        super().__init__()
        self._call_metadata = call_metadata
        self._client = client

    def sending_headers(self) -> Any:
        headers = {i[0]: i[1] for i in self._call_metadata}
        headers["x-request-id"] = str(uuid.uuid4())
        return headers


class _FlightCallMetadataMiddlewareFactory(fl.ClientMiddlewareFactory):
    def __init__(self, call_metadata: List[Tuple[str, str]], client: "DatasetClient"):
        super().__init__()
        self._call_metadata = call_metadata
        self._client = client

    def start_call(self, info: fl.CallInfo) -> Any:
        return _FlightCallMetadataMiddleware(self._call_metadata, self._client)


class DatasetClient:
    def __init__(self, address_and_port: str, access_token: str) -> None:
        self._endpoint = f"grpc+tls://{address_and_port}"
        self._ssl_config = create_grpc_ssl_config(
            address_and_port, do_verify_ssl=True, do_force_cadata_load=True
        )
        self._middleware = [
            _FlightCallMetadataMiddlewareFactory(
                [("authorization", f"Bearer {access_token}")], self
            )
        ]
        self._flight = fl.FlightClient(
            self._endpoint,
            override_hostname=f"dataset.{address_and_port.partition(':')[0]}",
            tls_root_certs=self._ssl_config.cadata,
            middleware=self._middleware,
        )

    def health_check(self) -> str:
        buf = pyarrow.allocate_buffer(0)
        action = fl.Action("HealthCheck", buf)
        result = next(self._flight.do_action(action))
        return result.body.to_pybytes().decode("utf-8")

    @_dataset_exception_handler
    def get_partitions_metadata(self, command: Command) -> List[PartitionMetadata]:
        descriptor = fl.FlightDescriptor.for_command(command.SerializeToString())
        flight_info = self._flight.get_flight_info(descriptor)
        partitions_metadata = []
        for endpoint in flight_info.endpoints:
            ticket = PartitionTicket()
            ticket.ParseFromString(endpoint.ticket.ticket)
            for location in endpoint.locations:
                partitions_metadata.append(
                    PartitionMetadata(
                        location=location.uri.decode("utf-8"),
                        format=ticket.format,
                        checksum=ticket.checksum,
                    )
                )
        return partitions_metadata

    @_dataset_exception_handler
    def fetch_partition(
        self,
        partition_metadata: PartitionMetadata,
        no_cache: bool = False,
        cache_dir: Optional[Path] = None,
    ) -> Partition:
        # support only parquet partitions
        if partition_metadata.format != PartitionTicket.Format.FORMAT_PARQUET:
            raise DatasetClientError(
                f"Unsupported partition format: {partition_metadata.format}"
            )
        # caching only possible if checksum available
        if len(partition_metadata.checksum) > 0 and not no_cache:
            cache = Cache(cache_dir=cache_dir).initialise()
            cache_path = cache.get_path_entry(partition_metadata.checksum)
            from_cache = cache_path is not None
            if not from_cache:
                download_path = cache.cache_dir.joinpath(str(uuid.uuid4()))
                urllib.request.urlretrieve(  # nosec urllib_urlopen
                    partition_metadata.location, filename=download_path
                )
                cache_path = cache.put_path_entry(
                    partition_metadata.checksum, download_path
                )
            return Partition(_read_parquet(cache_path), from_cache=from_cache)  # type: ignore
        # read directly from the remote location
        return Partition(_read_parquet(partition_metadata.location))

    @_dataset_exception_handler
    def get_dataset_writer(self, command: Command, schema: Any) -> Any:
        descriptor = fl.FlightDescriptor.for_command(command.SerializeToString())
        return self._flight.do_put(descriptor, schema)


def _read_parquet(path: Union[str, Path]) -> pandas.DataFrame:
    return pandas.read_parquet(path, engine="pyarrow")



import importlib.util
import os
import pickle  # nosec blacklist
import sys
from logging import Logger
from pathlib import Path
from typing import Any, Callable, Optional
from uuid import UUID

from layerapi.api.entity.model_train_status_pb2 import ModelTrainStatus
from layerapi.api.ids_pb2 import ModelTrainId

from layer.clients.model_catalog import ModelCatalogClient


def import_function(
    source_folder: Path, source_entrypoint: str, function_name: Optional[str] = None
) -> Callable[..., Any]:
    """
    Example:
    source_folder: ~/source
    source_entrypoint: models/survival_model/model.py
    function_name: train_model

    If the source_entrypoint is a pickle file, then the function will be deserialized from this file and
    the function_name parameter is unused and should be None.
    Otherwise, the soure_entrypoint.py is loaded as a Python module and the function_name function is
    extracted from that module and returned.
    """
    _, extension = os.path.splitext(source_entrypoint.replace(os.sep, "."))
    if extension.lower() == ".pickle" or extension.lower() == ".pkl":
        return _get_user_function_from_pickle_file(source_folder, source_entrypoint)
    else:
        assert function_name
        return _get_user_function_from_python_source(
            source_folder, source_entrypoint, function_name
        )


def _get_user_function_from_pickle_file(
    source_folder: Path, source_entrypoint: str
) -> Callable[..., Any]:
    expanded = os.path.expanduser(source_folder)
    full_entrypoint_path = os.path.join(expanded, source_entrypoint)
    with open(full_entrypoint_path, "rb") as f:
        return pickle.load(f)  # nosec pickle


def _get_user_function_from_python_source(
    source_folder: Path, source_entrypoint: str, function_name: str
) -> Callable[..., Any]:
    """
    Example input params:
    source_folder: ~/source
    source_entrypoint: models/survival_model/model.py
    function_name: train_model

    Then these vars would have these values:
    module_path: source.survival_model.model
    package: source.survival_model
    """
    expanded = os.path.expanduser(source_folder)
    source_dir = os.path.dirname(expanded)
    sys.path.append(source_dir)  # source becomes a top-level package
    module_path = (
        os.path.basename(source_folder)
        + "."
        + str(Path(source_entrypoint).with_suffix("")).replace(os.path.sep, ".")
    )
    package = (
        os.path.basename(source_folder)
        + "."
        + os.path.dirname(source_entrypoint).replace(os.path.sep, ".")
    )
    user_code_module = importlib.import_module(
        module_path,
        package=package,
    )
    sys.path.remove(source_dir)
    return getattr(user_code_module, function_name)


def update_train_status(
    model_catalog_client: ModelCatalogClient,
    train_id: UUID,
    train_status: "ModelTrainStatus.TrainStatus.V",
    logger: Logger,
    info: str = "",
) -> None:
    try:
        model_catalog_client.update_model_train_status(
            ModelTrainId(value=str(train_id)),
            ModelTrainStatus(train_status=train_status, info=info),
        )
    except Exception as e:
        reason = getattr(e, "message", repr(e))
        logger.error(
            f"Failure while trying to update the status of train ID {str(train_id)}: {reason}"
        )



import datetime
import json
import os
import uuid
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import grpc
from google.protobuf.json_format import MessageToDict
from grpc._cython.cygrpc import _Metadatum  # type: ignore

from layer.exceptions.exceptions import (
    LayerClientException,
    LayerClientResourceAlreadyExistsException,
    LayerClientResourceNotFoundException,
    LayerClientServiceUnavailableException,
    LayerClientTimeoutException,
    LayerResourceExhaustedException,
)
from layer.utils.session import _ENV_KEY_REQUEST_ID, UserSessionId, is_layer_debug_on


_OBFUSCATED_VALUE = "***"
_NON_OBFUSCATED_REQUEST_METADATA = [
    "x-request-id",
]


class GRPCErrorClientInterceptor(grpc.UnaryUnaryClientInterceptor):  # type: ignore
    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, Any], Any],
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        error: Optional[grpc.RpcError] = None
        try:
            outcome = continuation(client_call_details, request)
        except grpc.RpcError as err:
            error = err
        if outcome.exception() and isinstance(outcome.exception(), grpc.RpcError):
            error = outcome.exception()

        if error:
            if "Exception deserializing response" in str(error.details()):
                raise LayerClientException(
                    "Please make sure that you have the latest layer sdk version installed."
                )
            else:
                raise self._convert_rpc_error_to_client_exception(error)
        return outcome

    @staticmethod
    def _convert_rpc_error_to_client_exception(
        error: grpc.RpcError,
    ) -> Union[LayerClientException, grpc.RpcError]:
        error_details = str(error.details())
        request_id = ""
        for metadata in error.trailing_metadata():
            if metadata.key == "x-request-id":
                request_id = str(metadata.value)
                break
        client_error_message = f"code: {error.code().name}, details: {error_details}"
        if len(request_id) > 0:
            client_error_message = f"error id: {request_id}, {client_error_message}"

        if error.code() is grpc.StatusCode.DEADLINE_EXCEEDED:
            return LayerClientTimeoutException(client_error_message)
        elif error.code() is grpc.StatusCode.RESOURCE_EXHAUSTED:
            return LayerResourceExhaustedException(client_error_message)
        elif error.code() is grpc.StatusCode.UNAVAILABLE:
            return LayerClientServiceUnavailableException(error_details)
        elif error.code() is grpc.StatusCode.NOT_FOUND:
            return LayerClientResourceNotFoundException(error_details)
        elif error.code() is grpc.StatusCode.ALREADY_EXISTS:
            return LayerClientResourceAlreadyExistsException(error_details)
        return LayerClientException(client_error_message)


class LogRpcCallsInterceptor(grpc.UnaryUnaryClientInterceptor):  # type: ignore
    """
    _LogRpcCallsInterceptor will log all gRPC calls with obfuscated responses to
    the session logs.
    Metadata keys specified in _NON_OBFUSCATED_REQUEST_METADATA are not obfuscated.

    With LAYER_DEBUG environment variable set to truthy values ("True", "1"), responses
    will NOT be obfuscated.
    """

    _instance = None

    def __new__(cls, logs_file_path: Path) -> Any:
        if cls._instance is None:
            import atexit

            cls._instance = super(LogRpcCallsInterceptor, cls).__new__(cls)
            atexit.register(cls._instance.close)

            user_session_id = str(UserSessionId())
            cls._instance._user_session_id = user_session_id  # type: ignore
            cls.should_obfuscate_responses = not is_layer_debug_on()
            cls._instance._logs_file_path = logs_file_path  # type: ignore
            logs_file_path.parent.mkdir(parents=True, exist_ok=True)
            cls._instance._log_file = open(logs_file_path, "w")  # type: ignore
        elif logs_file_path != cls._instance._logs_file_path:  # type: ignore
            raise ValueError(
                "LogRpcCallsInterceptor instance already exists with different file path"
            )

        return cls._instance

    @classmethod
    def _clear_instance(cls) -> None:
        cls._instance = None

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, Any], Any],
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        started_utc = datetime.datetime.now(datetime.timezone.utc).timestamp()
        log_entry = {
            "started_utc": started_utc,
            "method": client_call_details.method,
            "request": MessageToDict(request),
        }

        if client_call_details.metadata is not None:
            obfuscated_metadata = self._obfuscate_metadata(client_call_details.metadata)
            log_entry["request_metadata"] = obfuscated_metadata

        try:
            outcome = continuation(client_call_details, request)
            if outcome.exception():
                err = outcome.exception()
                self._add_exception_to_log_entry(err, log_entry)
            if outcome.result():
                self._add_result_to_log_entry(outcome.result(), log_entry)
            log_entry["duration"] = (
                datetime.datetime.now(datetime.timezone.utc).timestamp() - started_utc
            )
            self._print_json_as_new_line(log_entry)
            return outcome
        except Exception as err:
            self._add_exception_to_log_entry(err, log_entry)
            log_entry["duration"] = (
                datetime.datetime.now(datetime.timezone.utc).timestamp() - started_utc
            )
            self._print_json_as_new_line(log_entry)
            raise err

    @staticmethod
    def _obfuscate_metadata(
        metadata: Tuple[Tuple[str, Union[str, bytes]], ...]
    ) -> Dict[str, Union[str, bytes]]:

        return {
            k: v if k in _NON_OBFUSCATED_REQUEST_METADATA else _OBFUSCATED_VALUE
            for k, v in metadata
        }

    @staticmethod
    def _obfuscate_trailing_metadata(
        metadata: Tuple[_Metadatum, ...]  # pytype: disable=invalid-annotation
    ) -> Dict[str, Union[str, bytes]]:
        return {
            m.key: m.value
            if m.key in _NON_OBFUSCATED_REQUEST_METADATA
            else _OBFUSCATED_VALUE
            for m in metadata
        }

    def _add_result_to_log_entry(self, result: Any, log_entry: Dict[str, Any]) -> None:
        result_dict = MessageToDict(result)
        if self.should_obfuscate_responses:
            log_entry["result"] = self._obfuscate_dict_values(result_dict)
        else:
            log_entry["result"] = result_dict

    @staticmethod
    def _obfuscate_dict_values(raw: Dict[str, Any]) -> Dict[str, Any]:
        out = {}
        for k, v in raw.items():
            out[k] = (
                LogRpcCallsInterceptor._obfuscate_dict_values(v)
                if isinstance(v, dict)
                else _OBFUSCATED_VALUE
            )
        return out

    def _add_exception_to_log_entry(
        self,
        exception: Exception,
        log_entry: Dict[str, Any],
    ) -> None:
        exception_as_dict = {
            "type": exception.__class__.__name__,
            "details": _OBFUSCATED_VALUE,
        }
        metadata: Optional[Dict[str, Any]] = None

        if isinstance(exception, grpc.RpcError):
            exception_as_dict[
                "code"
            ] = exception.code().name  # pytype: disable=attribute-error

            if not self.should_obfuscate_responses:
                exception_as_dict[
                    "details"
                ] = exception.details()  # pytype: disable=attribute-error

            if exception.trailing_metadata():  # pytype: disable=attribute-error
                metadata = LogRpcCallsInterceptor._obfuscate_trailing_metadata(
                    exception.trailing_metadata()  # pytype: disable=attribute-error
                )
        elif not self.should_obfuscate_responses:
            exception_as_dict["details"] = str(exception)

        log_entry["exception"] = exception_as_dict
        log_entry["result_metadata"] = metadata

    def _print_json_as_new_line(self, entry: Dict[str, Any]) -> None:
        print(json.dumps(entry), file=self._log_file)  # type: ignore
        self._flush()

    def close(self) -> None:
        if self._log_file:  # type: ignore
            self._log_file.close()  # type: ignore

    def _flush(self) -> None:
        if self._log_file:  # type: ignore
            self._log_file.flush()  # type: ignore


class RequestIdInterceptor(grpc.UnaryUnaryClientInterceptor):  # type: ignore
    _instance = None

    def __new__(cls) -> Any:
        if cls._instance is None:
            cls._instance = super(RequestIdInterceptor, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        super().__init__()
        # REQUEST_ID set by invoker if wants to use same request id for all requests
        request_id_candidate = os.getenv(_ENV_KEY_REQUEST_ID, default=None)
        if request_id_candidate is None:
            self._request_id = None
            return

        try:
            self._request_id = uuid.UUID(request_id_candidate)
        except Exception as e:
            raise ValueError(
                f"invalid {_ENV_KEY_REQUEST_ID} environment variable. Error message: {e}"
            )

    @classmethod
    def _clear_instance(cls) -> None:
        cls._instance = None

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, Any], Any],
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        metadata: List[Tuple[str, str]] = (
            client_call_details.metadata if client_call_details.metadata else []  # type: ignore
        )
        if self._request_id is not None:
            request_id = self._request_id
        else:
            request_id = uuid.uuid4()
        metadata.append(("x-request-id", str(request_id)))
        client_call_details = client_call_details._replace(metadata=metadata)  # type: ignore
        return continuation(client_call_details, request)



from pathlib import Path

import pandas as pd
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


class KerasModelFlavor(ModelFlavor):
    """An ML Model flavor implementation which handles persistence of Keras Models."""

    MODULE_KEYWORD = "keras"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_KERAS

    TOKENIZER_FILE = "tokenizer.pickle"

    def can_interpret_object(self, model_object: ModelObject) -> bool:
        # Check if model is a tensorflow keras
        try:
            import tensorflow.keras.models  # type: ignore

            if isinstance(model_object, tensorflow.keras.models.Model):
                return True
        except ImportError:
            pass

        # Check if model is a pure keras
        try:
            import keras.engine.network  # type: ignore

            if isinstance(model_object, keras.engine.network.Network):
                return True
        except ImportError:
            pass

        # Check if model is a tensorflow/pure keras tokenizer
        try:
            import keras

            if isinstance(model_object, keras.preprocessing.text.Tokenizer):
                return True
        except ImportError:
            pass

        return False

    def save_model_to_directory(
        self, model_object: ModelObject, directory: Path
    ) -> None:
        import cloudpickle  # type: ignore
        import keras
        import mlflow.keras

        if isinstance(model_object, keras.preprocessing.text.Tokenizer):
            directory.mkdir(parents=True, exist_ok=True)
            with open(directory / KerasModelFlavor.TOKENIZER_FILE, "wb") as handle:
                cloudpickle.dump(model_object, handle)
        else:
            mlflow.keras.save_model(model_object, path=directory.as_posix())

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        import mlflow.keras

        tokenizer_file = directory / KerasModelFlavor.TOKENIZER_FILE

        if tokenizer_file.exists():
            import cloudpickle

            with open(directory / KerasModelFlavor.TOKENIZER_FILE, "rb") as handle:
                model = cloudpickle.load(handle)
                return ModelRuntimeObjects(
                    model, lambda input_df: self.__predict(model, input_df)
                )
        else:
            model = mlflow.keras.load_model(directory.as_uri())
            return ModelRuntimeObjects(
                model, lambda input_df: self.__predict(model, input_df)
            )

    @staticmethod
    def __predict(model: ModelObject, input_df: pd.DataFrame) -> pd.DataFrame:
        #  https://www.tensorflow.org/api_docs/python/tf/keras/Model#predict
        predictions = model.predict(input_df)  # type: ignore
        return pd.DataFrame(predictions)



import typing
import uuid
from typing import Dict, List, Tuple

from layerapi.api.entity.history_event_pb2 import HistoryEvent
from layerapi.api.entity.run_metadata_pb2 import RunMetadata
from layerapi.api.entity.run_pb2 import Run as PBRun
from layerapi.api.entity.task_pb2 import Task as PBTask
from layerapi.api.ids_pb2 import ModelTrainId, ModelVersionId

from layer.clients.layer import LayerClient
from layer.contracts.projects import ApplyResult
from layer.contracts.runs import Run
from layer.exceptions.exceptions import (
    LayerClientTimeoutException,
    ProjectDatasetBuildExecutionException,
    ProjectModelExecutionException,
    ProjectRunnerError,
    ProjectRunTerminatedError,
)
from layer.exceptions.status_report import ExecutionStatusReportFactory
from layer.projects.utils import get_current_project_name
from layer.tracker.progress_tracker import RunProgressTracker
from layer.utils.session import is_layer_debug_on


class PollingStepFunction:
    def __init__(self, max_backoff_sec: float, backoff_multiplier: float):
        self._max_backoff = max_backoff_sec
        self._backoff_multiplier = backoff_multiplier

    def step(self, step: float) -> float:
        return min(self._max_backoff, step * self._backoff_multiplier)


_FormattedRunMetadata = Dict[Tuple["PBTask.Type.ValueType", str, str], str]


class ProgressTrackerUpdater:
    tracker: RunProgressTracker
    client: LayerClient
    apply_metadata: ApplyResult
    run: Run
    run_metadata: _FormattedRunMetadata

    def __init__(
        self,
        tracker: RunProgressTracker,
        apply_metadata: ApplyResult,
        run: Run,
        client: LayerClient,
    ):
        self.tracker = tracker
        self.apply_metadata = apply_metadata
        self.run = run
        self.client = client

    @staticmethod
    def _format_run_metadata(
        run_metadata: RunMetadata,
    ) -> _FormattedRunMetadata:
        return {
            (entry.task_type, entry.task_id, entry.key): entry.value
            for entry in list(run_metadata.entries)
        }

    def check_completion_and_update_tracker(
        self,
        response: typing.Union[
            Tuple[List[HistoryEvent], RunMetadata],
            LayerClientTimeoutException,
        ],
    ) -> bool:
        if isinstance(response, LayerClientTimeoutException):
            return False

        (events, run_metadata) = response
        self.run_metadata = self._format_run_metadata(run_metadata)

        for event in events:
            event_type = event.WhichOneof("event")

            if event_type == "run":
                run_status = event.run.run_status
                if run_status == PBRun.STATUS_TERMINATED:
                    raise ProjectRunTerminatedError(run_id=self.run.run_id)

                elif run_status == PBRun.STATUS_FAILED:
                    raise ProjectRunnerError("Run failed", self.run.run_id)

                elif run_status in [PBRun.STATUS_SUCCEEDED]:
                    return True

                elif run_status == PBRun.STATUS_INVALID:
                    # TODO: alert for this
                    _print_debug("Run status INVALID")

            if event_type == "task":
                self._update_tracker(event.task)

        return False

    def _update_tracker(self, task: PBTask) -> None:
        task_status = task.status
        if task_status == PBTask.STATUS_SCHEDULED:
            self._handle_task_scheduled(task)
        elif task_status == PBTask.STATUS_SUCCEEDED:
            self._handle_task_succeeded(task)
        elif task_status == PBTask.STATUS_FAILED:
            self._handle_task_failed(task)
        elif task_status == PBTask.STATUS_INVALID:
            # TODO: alert for this
            _print_debug("Task status INVALID")

    def _handle_task_succeeded(self, task: PBTask) -> None:
        task_id = task.id
        task_type = task.type
        if task_type == PBTask.TYPE_DATASET_BUILD:
            dataset_name = task_id
            dataset_path = f"{get_current_project_name()}/datasets/{dataset_name}"
            dataset_build_id = uuid.UUID(
                self.run_metadata[(task_type, dataset_path, "build-id")]
            )
            dataset = self.client.data_catalog.get_dataset_by_build_id(dataset_build_id)
            self.tracker.mark_dataset_built(
                name=dataset_name,
                version=dataset.version,
                build_index=dataset.build.index,
            )
        elif task_type == PBTask.TYPE_MODEL_TRAIN:
            train_id = uuid.UUID(self.run_metadata[(task_type, task_id, "train-id")])
            model_name = self._find_model_name_by_version_id(task_id)
            train_index = self.client.model_catalog.get_model_train(
                ModelTrainId(value=str(train_id))
            ).index
            version_name = self.client.model_catalog.get_model_version(
                ModelVersionId(value=str(task_id))
            ).name
            self.tracker.mark_model_trained(
                name=model_name,
                train_index=str(train_index),
                version=version_name,
            )
        else:
            # TODO: alert for this
            _print_debug(f"Task type not handled {task_type}")

    def _handle_task_failed(self, task: PBTask) -> None:
        assert self.run.run_id
        task_id = task.id
        task_type = task.type
        task_info = task.info
        if task_type == PBTask.TYPE_DATASET_BUILD:
            exc_ds = ProjectDatasetBuildExecutionException(
                self.run.run_id,
                task_id,
                ExecutionStatusReportFactory.from_json(task_info),
            )
            self.tracker.mark_dataset_failed(name=task_id, reason=exc_ds.message)
            raise exc_ds
        elif task_type == PBTask.TYPE_MODEL_TRAIN:
            exc_model = ProjectModelExecutionException(
                self.run.run_id,
                task_id,
                ExecutionStatusReportFactory.from_json(task_info),
            )
            self.tracker.mark_model_train_failed(
                name=self._find_model_name_by_version_id(task_id),
                reason=f"{exc_model.message}",
            )
            raise exc_model
        else:
            # TODO: alert for this
            _print_debug(f"Task type not handled {task_type}")

    def _handle_task_scheduled(self, task: PBTask) -> None:
        task_id = task.id
        task_type = task.type
        if task_type == PBTask.TYPE_DATASET_BUILD:
            self.tracker.mark_dataset_building(name=task_id)
        elif task_type == PBTask.TYPE_MODEL_TRAIN:
            self.tracker.mark_model_training(
                name=self._find_model_name_by_version_id(task_id)
            )
        else:
            # TODO: alert for this
            _print_debug(f"Task type not handled {task_type}")

    def _find_model_name_by_version_id(self, version_id: str) -> str:
        for definition in self.run.definitions:
            if str(definition.version_id) == version_id:
                return definition.asset_name
        raise KeyError(version_id)


def _print_debug(msg: str) -> None:
    if is_layer_debug_on():
        print(msg)



from .dataset_decorator import dataset  # noqa
from .fabric_decorator import fabric  # noqa
from .model_decorator import model  # noqa
from .pip_requirements_decorator import pip_requirements  # noqa
from .resources_decorator import resources  # noqa



import enum
import uuid
from dataclasses import dataclass, field
from typing import Any, Callable, List, Mapping, Optional, Sequence, Union

import pandas

from .asset import AssetPath, AssetType, BaseAsset


def _create_empty_data_frame() -> "pandas.DataFrame":
    return pandas.DataFrame()


@enum.unique
class DatasetBuildStatus(enum.IntEnum):
    INVALID = 0
    STARTED = 1
    COMPLETED = 2
    FAILED = 3


@dataclass(frozen=True)
class DatasetBuild:
    id: uuid.UUID = field(default_factory=uuid.uuid4)
    status: DatasetBuildStatus = DatasetBuildStatus.INVALID
    info: str = ""
    index: str = ""


class Dataset(BaseAsset):
    """
    Provides access to datasets defined in Layer.

    You can retrieve an instance of this object with :code:`layer.get_dataset()`.

    This class should not be initialized by end-users.

    .. code-block:: python

        # Fetches the `titanic` dataset
        Dataset("titanic")

    """

    def __init__(
        self,
        asset_path: Union[str, AssetPath],
        id: Optional[uuid.UUID] = None,
        dependencies: Optional[Sequence[BaseAsset]] = None,
        version_id: Optional[uuid.UUID] = None,
        description: str = "",
        uri: str = "",
        schema: str = "{}",
        metadata: Optional[Mapping[str, str]] = None,
        build: Optional[DatasetBuild] = None,
        # https://stackoverflow.com/questions/51811024/mypy-type-checking-on-callable-thinks-that-member-variable-is-a-method
        _pandas_df_factory: Optional[Callable[[], "pandas.DataFrame"]] = None,
        version: Optional[str] = None,  # TODO(volkan) is this needed?
    ):
        super().__init__(
            path=asset_path,
            id=id,
            dependencies=dependencies,
        )
        self._version_id = version_id
        self.description = description
        self.schema = schema
        self.metadata = metadata if metadata is not None else {}
        self.build = build or DatasetBuild()
        self.__pandas_df_factory = _pandas_df_factory or _create_empty_data_frame
        self.uri = uri
        self.version = version if version is not None else ""

    @property
    def asset_type(self) -> AssetType:
        return AssetType.DATASET

    def _pandas_df_factory(self) -> "pandas.DataFrame":
        assert self.__pandas_df_factory
        return self.__pandas_df_factory()

    @property
    def is_build_completed(self) -> bool:
        return self.build.status == DatasetBuildStatus.COMPLETED

    @property
    def build_info(self) -> str:
        return self.build.info

    def to_pandas(self) -> "pandas.DataFrame":
        """
        Fetches the dataset as a Pandas dataframe.

        :return: A Pandas dataframe containing your dataset.
        """
        return self._pandas_df_factory()

    def to_pytorch(
        self,
        transformer: Callable[[Any], Any],
        tensors: Optional[List[str]] = None,
        batch_size: Optional[int] = 1,
        shuffle: bool = False,
        sampler: Any = None,
        batch_sampler: Any = None,
        num_workers: int = 0,
        collate_fn: Optional[Callable[[Any], Any]] = None,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0,
        worker_init_fn: Optional[Callable[[int], None]] = None,
        prefetch_factor: int = 2,
        persistent_workers: bool = False,
        generator: Any = None,
    ) -> Any:
        """
        Fetches the dataset as a Pytorch DataLoader.
        :param transformer: Function to apply the transformations to the data
        :param tensors: List of columns to fetch
        :param batch_size: how many samples per batch to load (default: 1).
        :param shuffle: set to True to have the data reshuffled at every epoch (default: False).
        :param sampler: defines the strategy to draw samples from the dataset. Can be any Iterable with __len__ implemented. If specified, shuffle must not be specified.
        :param batch_sampler: like sampler, but returns a batch of indices at a time. Mutually exclusive with batch_size, shuffle, sampler, and drop_last.
        :param num_workers: how many subprocesses to use for data loading. 0 means that the data will be loaded in the main process. (default: 0)
        :param collate_fn: merges a list of samples to form a mini-batch of Tensor(s). Used when using batched loading from a map-style dataset.
        :param pin_memory: If True, the data loader will copy Tensors into CUDA pinned memory before returning them. If your data elements are a custom type, or your collate_fn returns a batch that is a custom type, see the example below.
        :param drop_last: set to True to drop the last incomplete batch, if the dataset size is not divisible by the batch size. If False and the size of dataset is not divisible by the batch size, then the last batch will be smaller. (default: False)
        :param timeout:  if positive, the timeout value for collecting a batch from workers. Should always be non-negative. (default: 0)
        :param worker_init_fn: If not None, this will be called on each worker subprocess with the worker id (an int in [0, num_workers - 1]) as input, after seeding and before data loading. (default: None)
        :param prefetch_factor: Number of samples loaded in advance by each worker. 2 means there will be a total of 2 * num_workers samples prefetched across all workers. (default: 2)
        :param persistent_workers: If True, the data loader will not shutdown the worker processes after a dataset has been consumed once. This allows to maintain the workers Dataset instances alive. (default: False)
        :param generator:  If not None, this RNG will be used by RandomSampler to generate random indexes and multiprocessing to generate base_seed for workers. (default: None)
        :return: torch.utils.data.DataLoader
        """

        # Check if `torch` is installed
        try:
            import torch
        except ImportError:
            raise Exception(
                "PyTorch needs to be installed to run `to_pytorch()`. Try `pip install torch`"
            )

        class PytorchDataset(torch.utils.data.Dataset[Any]):
            # TODO: Streaming data fetching for faster data access

            def __init__(self, df: pandas.DataFrame, transformer: Callable[[Any], Any]):
                self.df = df
                self.transformer = transformer

            def __getitem__(self, key: Union[slice, int]) -> Any:
                if isinstance(key, slice):
                    # get the start, stop, and step from the slice
                    return [self[ii] for ii in range(*key.indices(len(self)))]
                elif isinstance(key, int):
                    # handle negative indices
                    if key < 0:
                        key += len(self)
                    if key < 0 or key >= len(self):
                        raise IndexError("The index (%d) is out of range." % key)
                    # get the data from direct index
                    return self.transformer(self.df.iloc[key])
                else:
                    raise TypeError("Invalid argument type.")

            def __len__(self) -> int:
                return self.df.shape[0]

        df = self.to_pandas()
        if tensors:
            df = df[tensors]
        dataset = PytorchDataset(df, transformer)
        return torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            generator=generator,
            prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
        )


@dataclass(frozen=True)
class SortField:
    name: str
    descending: bool = False



import tempfile
import uuid
from contextlib import contextmanager
from logging import Logger
from pathlib import Path
from typing import Iterator

import polling  # type: ignore
from layerapi.api.entity.model_train_status_pb2 import (
    ModelTrainStatus as PBModelTrainStatus,
)
from layerapi.api.entity.model_version_pb2 import ModelVersion
from layerapi.api.ids_pb2 import ModelTrainId
from layerapi.api.service.modeltraining.model_training_api_pb2 import (
    GetModelTrainStatusRequest,
    GetSourceCodeUploadCredentialsRequest,
    GetSourceCodeUploadCredentialsResponse,
    StartModelTrainingRequest,
)
from layerapi.api.service.modeltraining.model_training_api_pb2_grpc import (
    ModelTrainingAPIStub,
)
from layerapi.api.value.s3_path_pb2 import S3Path

from layer.config import ClientConfig
from layer.contracts.runs import FunctionDefinition
from layer.exceptions.exceptions import LayerClientException
from layer.utils.file_utils import tar_directory
from layer.utils.grpc import create_grpc_channel
from layer.utils.s3 import S3Util


class ModelTrainingClient:
    _service: ModelTrainingAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
    ):
        self._config = config.model_training
        self._logger = logger
        self._access_token = config.access_token
        self._s3_endpoint_url = config.s3.endpoint_url
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

    @contextmanager
    def init(self) -> Iterator["ModelTrainingClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = ModelTrainingAPIStub(channel=channel)
            yield self

    def upload_training_files(
        self, model: FunctionDefinition, model_version_id: uuid.UUID
    ) -> S3Path:
        response = self.get_source_code_upload_credentials(model_version_id)

        with tempfile.TemporaryDirectory() as tmp_dir:
            tar_directory(f"{tmp_dir}/{model.asset_name}.tgz", model.function_home_dir)
            S3Util.upload_dir(
                Path(tmp_dir),
                response.credentials,
                response.s3_path,
                endpoint_url=self._s3_endpoint_url,
            )

        return response.s3_path

    def get_source_code_upload_credentials(
        self, model_version_id: uuid.UUID
    ) -> GetSourceCodeUploadCredentialsResponse:
        return self._service.GetSourceCodeUploadCredentials(
            GetSourceCodeUploadCredentialsRequest(source_name=str(model_version_id))
        )

    def train_model(
        self,
        version: ModelVersion,
    ) -> uuid.UUID:
        response: GetSourceCodeUploadCredentialsResponse = (
            self.get_source_code_upload_credentials(uuid.UUID(version.id.value))
        )
        self._logger.debug(
            f"GetSourceCodeUploadCredentialsResponse response: {str(response)}"
        )
        return self._execute_regular_train(train_id=version.latest_train_id)

    def _execute_regular_train(self, train_id: ModelTrainId) -> uuid.UUID:
        request: StartModelTrainingRequest = StartModelTrainingRequest(
            model_train_id=train_id
        )
        self._logger.debug(f"StartExecuteModelTrainRequest request: {str(request)}")
        train_response = self._service.StartModelTraining(request)

        def is_train_completed(status: PBModelTrainStatus) -> bool:
            return (
                status.train_status == PBModelTrainStatus.TRAIN_STATUS_SUCCESSFUL
                or status.train_status == PBModelTrainStatus.TRAIN_STATUS_FAILED
            )

        polling.poll(
            lambda: self._get_model_train_status(train_response.id.value),
            check_success=is_train_completed,
            step=5,
            poll_forever=True,
        )
        status = self._get_model_train_status(train_response.id.value)
        if status.train_status == PBModelTrainStatus.TRAIN_STATUS_FAILED:
            raise LayerClientException(f"regular train failed. Info: {status.info}")
        return uuid.UUID(train_response.id.value)

    def _get_model_train_status(self, id: uuid.UUID) -> PBModelTrainStatus:
        response = self._service.GetModelTrainStatus(
            GetModelTrainStatusRequest(id=ModelTrainId(value=str(id)))
        )
        return response.train_status



import hashlib
from typing import List
from uuid import UUID

from layer import current_project_name
from layer.clients.layer import LayerClient
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.projects import Project
from layer.contracts.runs import FunctionDefinition
from layer.exceptions.exceptions import ProjectInitializationException
from layer.global_context import current_account_name


def verify_project_exists_and_retrieve_project_id(
    client: LayerClient, project_full_name: ProjectFullName
) -> UUID:
    project = client.project_service_client.get_project(project_full_name)
    if not project:
        raise ProjectInitializationException(
            f"Project '{project_full_name.path}' does not exist."
        )
    return project.id


def get_or_create_remote_project(
    client: LayerClient, project_full_name: ProjectFullName
) -> Project:
    project = client.project_service_client.get_project(project_full_name)
    if project is not None:
        return project

    return client.project_service_client.create_project(
        project_full_name,
    )


def get_current_project_name() -> str:
    current_project_name_ = current_project_name()
    if not current_project_name_:
        raise ProjectInitializationException(
            "Please specify the current project name globally with `layer.init('project-name')`"
        )
    return current_project_name_


def get_current_project_full_name() -> ProjectFullName:
    current_account_name_ = current_account_name()
    current_project_name_ = current_project_name()
    if not current_account_name_ or not current_project_name_:
        raise ProjectInitializationException(
            "Please specify the current project name globally with `layer.init('project-name')`"
        )
    return ProjectFullName(
        account_name=current_account_name_,
        project_name=current_project_name_,
    )


def calculate_hash_by_definitions(definitions: List[FunctionDefinition]) -> str:
    files_hash = hashlib.sha256()

    for definition in definitions:
        files_hash.update(definition.get_pickled_function())

    return files_hash.hexdigest()



import time
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum, unique
from typing import Any, Dict, Optional, Union

from yarl import URL

from layer.contracts.assets import AssetType


class ResourceTransferState:
    def __init__(self, name: Optional[str] = None) -> None:
        self._name: Optional[str] = name
        self._total_num_files: int = 0
        self._transferred_num_files: int = 0
        self._total_resource_size_bytes: int = 0
        self._transferred_resource_size_bytes: int = 0
        self._timestamp_to_bytes_sent: Dict[Any, int] = defaultdict(int)
        self._how_many_previous_seconds = 20

    def increment_num_transferred_files(self, increment_with: int) -> None:
        self._transferred_num_files += increment_with

    def increment_transferred_resource_size_bytes(self, increment_with: int) -> None:
        self._transferred_resource_size_bytes += increment_with
        self._timestamp_to_bytes_sent[self._get_current_timestamp()] += increment_with

    def get_bandwidth_in_previous_seconds(self) -> int:
        curr_timestamp = self._get_current_timestamp()
        valid_seconds = 0
        bytes_transferred = 0
        # Exclude start and current sec as otherwise result would not account for data that will be sent
        # within current second after the call of this func
        for i in range(1, self._how_many_previous_seconds + 1):
            look_at = curr_timestamp - i
            if look_at >= 0 and look_at in self._timestamp_to_bytes_sent:
                valid_seconds += 1
                bytes_transferred += self._timestamp_to_bytes_sent[look_at]
        if valid_seconds == 0:
            return 0
        return round(bytes_transferred / valid_seconds)

    def get_eta_seconds(self) -> int:
        bandwidth = self.get_bandwidth_in_previous_seconds()
        if bandwidth == 0:
            return 0
        return round(
            (self._total_resource_size_bytes - self._transferred_resource_size_bytes)
            / bandwidth
        )

    @staticmethod
    def _get_current_timestamp() -> int:
        return round(time.time())

    @property
    def transferred_num_files(self) -> int:
        return self._transferred_num_files

    @property
    def total_num_files(self) -> int:
        return self._total_num_files

    @total_num_files.setter
    def total_num_files(self, value: int) -> None:
        self._total_num_files = value

    @property
    def total_resource_size_bytes(self) -> int:
        return self._total_resource_size_bytes

    @total_resource_size_bytes.setter
    def total_resource_size_bytes(self, value: int) -> None:
        self._total_resource_size_bytes = value

    @property
    def transferred_resource_size_bytes(self) -> int:
        return self._transferred_resource_size_bytes

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    def __str__(self) -> str:
        return str(self.__dict__)


class DatasetTransferState:
    def __init__(self, total_num_rows: int, name: Optional[str] = None) -> None:
        self._name: Optional[str] = name
        self._total_num_rows: int = total_num_rows
        self._transferred_num_rows: int = 0
        self._timestamp_to_rows_sent: Dict[Any, int] = defaultdict(int)
        self._how_many_previous_seconds = 20

    def increment_num_transferred_rows(self, increment_with: int) -> None:
        self._transferred_num_rows += increment_with
        self._timestamp_to_rows_sent[self._get_current_timestamp()] += increment_with

    def _get_rows_per_sec(self) -> int:
        curr_timestamp = self._get_current_timestamp()
        valid_seconds = 0
        rows_transferred = 0
        # Exclude start and current sec as otherwise result would not account for data that will be sent
        # within current second after the call of this func
        for i in range(1, self._how_many_previous_seconds + 1):
            look_at = curr_timestamp - i
            if look_at >= 0 and look_at in self._timestamp_to_rows_sent:
                valid_seconds += 1
                rows_transferred += self._timestamp_to_rows_sent[look_at]
        if valid_seconds == 0:
            return 0
        return round(rows_transferred / valid_seconds)

    def get_eta_seconds(self) -> int:
        rows_per_sec = self._get_rows_per_sec()
        if rows_per_sec == 0:
            return 0
        return round((self._total_num_rows - self._transferred_num_rows) / rows_per_sec)

    @staticmethod
    def _get_current_timestamp() -> int:
        return round(time.time())

    @property
    def transferred_num_rows(self) -> int:
        return self._transferred_num_rows

    @property
    def total_num_rows(self) -> int:
        return self._total_num_rows

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    def __str__(self) -> str:
        return str(self.__dict__)


@unique
class AssetTrackerStatus(str, Enum):
    PENDING = "pending"
    SAVING = "saving"
    BUILDING = "building"
    TRAINING = "training"
    DONE = "done"
    ERROR = "error"
    ASSERTING = "asserting"
    ASSERTED = "asserted"
    RESOURCE_UPLOADING = "uploading resources"
    RESULT_UPLOADING = "uploading result"
    ASSET_DOWNLOADING = "downloading asset"
    ASSET_FROM_CACHE = "asset from cache"
    ASSET_LOADED = "asset loaded"

    @property
    def is_running(self) -> bool:
        return self in (
            AssetTrackerStatus.SAVING,
            AssetTrackerStatus.BUILDING,
            AssetTrackerStatus.TRAINING,
            AssetTrackerStatus.ASSERTING,
            AssetTrackerStatus.RESOURCE_UPLOADING,
            AssetTrackerStatus.RESULT_UPLOADING,
            AssetTrackerStatus.ASSET_DOWNLOADING,
            AssetTrackerStatus.ASSET_FROM_CACHE,
        )

    @property
    def is_finished(self) -> bool:
        return self in (
            AssetTrackerStatus.DONE,
            AssetTrackerStatus.ERROR,
            AssetTrackerStatus.ASSET_LOADED,
        )


@dataclass
class AssetTracker:
    type: AssetType
    name: str
    status: AssetTrackerStatus = AssetTrackerStatus.PENDING
    base_url: Optional[URL] = None
    error_reason: str = ""
    resource_transfer_state: Optional[ResourceTransferState] = None
    dataset_transfer_state: Optional[DatasetTransferState] = None
    model_transfer_state: Optional[ResourceTransferState] = None
    asset_download_transfer_state: Optional[
        Union[ResourceTransferState, DatasetTransferState]
    ] = None
    loading_cache_asset: Optional[str] = None
    version: Optional[str] = None
    build_idx: Optional[str] = None

    @property
    def is_running(self) -> bool:
        return self.status.is_running

    @property
    def is_finished(self) -> bool:
        return self.status.is_finished



from pathlib import Path

import pandas as pd
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


class PyTorchModelFlavor(ModelFlavor):
    """An ML Model flavor implementation which handles persistence of PyTorch Models."""

    MODULE_KEYWORD = "torch"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_PYTORCH

    def save_model_to_directory(
        self, model_object: ModelObject, directory: Path
    ) -> None:
        import mlflow.pytorch

        models_module_path = None

        # Check if the model is a YOLOV5 model
        if model_object.__class__.__name__ == "AutoShape":
            try:
                try:
                    # This is `models` module created cloning the official YoloV5 repo
                    import models  # type: ignore
                    import utils  # type: ignore
                except ModuleNotFoundError:
                    # Fallback: This is `models` module created installing the yolov5 from pypi
                    from yolov5 import models, utils  # type: ignore

                # YOLO models has a wrapper around the pytorch model.
                # We pack the required wrapper modules with the model.
                models_module_path = [
                    Path(models.__file__).parent,
                    Path(utils.__file__).parent,
                ]
            except ModuleNotFoundError:
                raise Exception("Can't save YOLO model. `models` module not found!")

        mlflow.pytorch.save_model(
            model_object, path=directory.as_posix(), code_paths=models_module_path
        )

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        import mlflow.pytorch
        import torch

        model = mlflow.pytorch.load_model(
            directory.as_uri(),
            map_location=torch.device("cpu"),
        )
        return ModelRuntimeObjects(
            model, lambda input_df: self.__predict(model, input_df)
        )

    @staticmethod
    def __predict(model: ModelObject, input_df: pd.DataFrame) -> pd.DataFrame:
        from mlflow.pytorch import _PyTorchWrapper

        model = _PyTorchWrapper(model)
        return model.predict(input_df)  # type: ignore



import time
import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Callable, List, Tuple

from layerapi.api.entity.user_log_line_pb2 import UserLogLine as PBUserLogLine
from layerapi.api.ids_pb2 import RunId

from layer.clients.layer import LayerClient
from layer.contracts.assets import AssetType


POLLING_INTERVAL_SEC = 3
LOGS_BUFFER_INTERVAL = 20  # Minimum amount of time to have as a buffer for logs


@dataclass(frozen=True)
class UserLogLine:
    asset_name: str = field(default_factory=str)
    type: str = field(default_factory=str)
    host_name: str = field(default_factory=str)
    time: datetime = datetime.utcfromtimestamp(0)
    log: str = field(default_factory=str)


def __convert_log_line(line: PBUserLogLine) -> UserLogLine:
    return UserLogLine(
        type=__convert_entity(line.type).name,
        asset_name=line.entity_name,
        host_name=line.host_name,
        time=line.time.ToDatetime(),
        log=line.log,
    )


def __convert_entity(entity: "PBUserLogLine.TaskType.ValueType") -> AssetType:
    if entity == PBUserLogLine.TASK_TYPE_MODEL_TRAIN:
        return AssetType.MODEL
    elif entity == PBUserLogLine.TASK_TYPE_DATASET_BUILD:
        return AssetType.DATASET
    else:
        raise Exception(f"Unable to convert {entity}")


def __get_lines(
    pipeline_run_id: uuid.UUID, continuation_token: str, client: LayerClient
) -> Tuple[List[UserLogLine], str]:
    proto_lines, continuation_token = client.user_logs.get_pipeline_run_logs(
        pipeline_run_id, continuation_token
    )
    converted_lines = [__convert_log_line(line) for line in proto_lines]
    return converted_lines, continuation_token


def __format_line(line: UserLogLine) -> str:
    time_str = line.time.strftime("%H:%M:%S")
    return f"\033[0;33m{time_str} \033[1;32m{line.asset_name}\033[0m: {line.log}"


def show_pipeline_run_logs(
    client: LayerClient,
    pipeline_run_id: str,
    follow: bool,
    polling_interval_sec: float = POLLING_INTERVAL_SEC,
    printer: Callable[[str], Any] = print,
    evaluate_until_callback: Callable[[], bool] = lambda: True,
) -> None:
    curr_token = ""
    while evaluate_until_callback():
        lines, next_token = __get_lines(uuid.UUID(pipeline_run_id), curr_token, client)
        for line in lines:
            formatted = __format_line(line)
            printer(formatted)
        prev_token = curr_token
        curr_token = next_token
        if prev_token == curr_token and follow is False:
            break
        time.sleep(polling_interval_sec)


def get_pipeline_run_logs(
    client: LayerClient,
    pipeline_run_id: RunId,
    polling_interval_sec: float = POLLING_INTERVAL_SEC,
) -> List[UserLogLine]:
    lines: List[UserLogLine] = []
    curr_token = ""
    while True:
        curr_lines, next_token = __get_lines(
            uuid.UUID(pipeline_run_id.value), curr_token, client
        )
        lines.extend(curr_lines)
        prev_token = curr_token
        curr_token = next_token
        if prev_token == curr_token:
            break
        time.sleep(polling_interval_sec)
    return lines



from contextlib import contextmanager
from logging import Logger
from typing import Iterator, List, Optional, Tuple

from layerapi.api.entity.history_event_pb2 import HistoryEvent
from layerapi.api.entity.operations_pb2 import ExecutionPlan
from layerapi.api.entity.run_filter_pb2 import RunFilter
from layerapi.api.entity.run_metadata_pb2 import RunMetadata
from layerapi.api.entity.run_pb2 import Run
from layerapi.api.ids_pb2 import RunId
from layerapi.api.service.flowmanager.flow_manager_api_pb2 import (
    GetRunByIdRequest,
    GetRunHistoryAndMetadataRequest,
    GetRunsRequest,
    StartRunV2Request,
    TerminateRunRequest,
)
from layerapi.api.service.flowmanager.flow_manager_api_pb2_grpc import (
    FlowManagerAPIStub,
)
from layerapi.api.value.sha256_pb2 import Sha256

from layer.config import ClientConfig
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.runs import GetRunsFunction
from layer.utils.grpc import create_grpc_channel


class FlowManagerClient:
    _service: FlowManagerAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
    ):
        self._config = config.flow_manager
        self._logger = logger
        self._access_token = config.access_token
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

    @contextmanager
    def init(self) -> Iterator["FlowManagerClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = FlowManagerAPIStub(channel=channel)
            yield self

    def start_run(
        self,
        project_full_name: ProjectFullName,
        execution_plan: ExecutionPlan,
        project_files_hash: str,
        user_command: str,
    ) -> RunId:
        response = self._service.StartRunV2(
            request=StartRunV2Request(
                project_full_name=project_full_name.path,
                plan=execution_plan,
                project_files_hash=Sha256(value=project_files_hash),
                user_command=user_command,
            )
        )
        return response.run_id

    def get_run(self, run_id: RunId) -> Run:
        response = self._service.GetRunById(GetRunByIdRequest(run_id=run_id))
        return response.run

    def get_run_status_history_and_metadata(
        self, run_id: RunId
    ) -> Tuple[List[HistoryEvent], RunMetadata]:
        response = self._service.GetRunHistoryAndMetadata(
            GetRunHistoryAndMetadataRequest(run_id=run_id)
        )
        return list(response.events), response.run_metadata

    def get_runs(
        self,
        pipeline_run_id: str,
        list_all: bool,
        project_name: Optional[str] = None,
    ) -> Tuple[List[Run], GetRunsFunction]:
        def get_runs_func() -> List[Run]:
            if pipeline_run_id:
                run = self._service.GetRunById(
                    request=GetRunByIdRequest(run_id=RunId(value=pipeline_run_id))
                ).run
                return [run]
            else:
                request = GetRunsRequest(
                    filters=self._get_run_filters(list_all, project_name=project_name)
                )
                _runs = self._service.GetRuns(request=request).runs
                return list(_runs)

        runs = get_runs_func()
        return runs, get_runs_func

    @staticmethod
    def _get_run_filters(
        list_all: bool, project_name: Optional[str] = None
    ) -> List[RunFilter]:
        filters = []
        if not list_all:
            filters.append(
                RunFilter(
                    status_filter=RunFilter.StatusFilter(
                        run_statuses=[
                            Run.Status.STATUS_RUNNING,
                            Run.Status.STATUS_SCHEDULED,
                        ]
                    )
                )
            )
        if project_name:
            filters.append(
                RunFilter(
                    project_name_filter=RunFilter.ProjectNameFilter(
                        project_name=project_name
                    )
                )
            )
        return filters

    def terminate_run(self, run_id: RunId) -> Run:
        response = self._service.TerminateRun(TerminateRunRequest(run_id=run_id))
        return response.run_id



import asyncio
import logging
import os
import urllib
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from types import SimpleNamespace
from typing import Any
from urllib.parse import urlparse

import aiohttp

from layer.clients.layer import LayerClient
from layer.contracts.assets import AssetType
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.runs import FunctionDefinition, ResourcePath, Run
from layer.contracts.tracker import ResourceTransferState
from layer.tracker.progress_tracker import RunProgressTracker


def _strip_resource_root_path(path: str) -> str:
    """
    For the resource paths of the form /<organisation_id>/<project>/resources/<function>/<subdir1>/...,
    return relative resource paths. For the example above, the result would be <subdir1>/...
    :param path: resource path.
    :return: relative function resource path.
    """
    unquoted_path = urllib.parse.unquote(path)
    n = 0
    for i, c in enumerate(unquoted_path):
        if c == "/":
            n += 1
            if n == 5:
                return unquoted_path[(i + 1) :]
    return unquoted_path


class ResourceManager:
    def __init__(self, client: LayerClient):
        self._client = client
        self._log = logging.getLogger(__name__)

    async def _upload_resource(
        self,
        project_full_name: ProjectFullName,
        function_name: str,
        resource_path: ResourcePath,
        file_path: str,
        session: aiohttp.ClientSession,
        state: ResourceTransferState,
    ) -> None:
        """
        Upload single local resource file.
        """
        self._log.debug(
            f"Upload resource file {file_path} of function {function_name} to {resource_path.path}"
        )
        resource_uri = self._client.data_catalog.get_resource_paths(
            project_full_name=project_full_name,
            function_name=function_name,
            path=resource_path.path,
        )[0]
        with open(file_path, mode="rb") as f:
            await session.put(
                resource_uri,
                data=f,
                timeout=aiohttp.ClientTimeout(total=None),
                trace_request_ctx={"state": state},
            )
        state.increment_num_transferred_files(1)

    def _update_resource_paths_index(self, run: Run) -> None:
        for function in run.definitions:
            self._client.data_catalog.update_resource_paths_index(
                project_full_name=run.project_full_name,
                function_name=function.func_name,
                paths=[
                    local_path
                    for resource_path in function.resource_paths
                    for local_path in resource_path.local_relative_paths()
                ],
            )

    @staticmethod
    async def _on_request_chunk_sent(
        session: aiohttp.ClientSession,
        ctx: SimpleNamespace,
        params: aiohttp.TraceRequestChunkSentParams,
    ) -> None:
        state: ResourceTransferState = ctx.trace_request_ctx["state"]
        state.increment_transferred_resource_size_bytes(len(params.chunk))

    async def _upload_resources(self, run: Run, tracker: RunProgressTracker) -> None:
        """
        Collect and upload local files as resources for all functions decorated with `@resource`.
        """
        self._update_resource_paths_index(run)
        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_chunk_sent.append(self._on_request_chunk_sent)
        async with aiohttp.ClientSession(
            raise_for_status=True, trace_configs=[trace_config]
        ) as session:
            upload_tasks = [
                self._upload_resources_for_function(run, tracker, function, session)
                for function in run.definitions
            ]
            await asyncio.gather(*upload_tasks)

    async def _upload_resources_for_function(
        self,
        run: Run,
        tracker: RunProgressTracker,
        function: FunctionDefinition,
        session: aiohttp.ClientSession,
    ) -> None:
        state = ResourceTransferState()
        upload_tasks = []
        total_num_files = 0
        total_files_size_bytes = 0
        for resource_path in function.resource_paths:
            for local_path in resource_path.local_relative_paths():
                total_num_files += 1
                total_files_size_bytes += os.path.getsize(os.path.abspath(local_path))
                upload_task = self._upload_resource(
                    project_full_name=run.project_full_name,
                    function_name=function.func_name,
                    resource_path=ResourcePath(path=local_path),
                    file_path=os.path.abspath(local_path),
                    session=session,
                    state=state,
                )
                upload_tasks.append(upload_task)
        state.total_num_files = total_num_files
        state.total_resource_size_bytes = total_files_size_bytes

        asset_type = function.asset_type
        asset_name = function.asset_name
        if asset_type == AssetType.DATASET:
            tracker.mark_dataset_resources_uploading(asset_name, state)
        elif asset_type == AssetType.MODEL:
            tracker.mark_model_resources_uploading(asset_name, state)
        await asyncio.gather(*upload_tasks)
        if asset_type == AssetType.DATASET:
            tracker.mark_dataset_resources_uploaded(asset_name)
        elif asset_type == AssetType.MODEL:
            tracker.mark_model_resources_uploaded(asset_name)

    def wait_resource_upload(self, run: Run, tracker: RunProgressTracker) -> None:
        """
        Collect and upload local files as resources for all functions decorated with `@resources`.
        """
        self._log.debug("Uploading resources")
        # run in a separate thread to make sure no other loop is running
        with ThreadPoolExecutor(max_workers=1) as executor:
            upload = executor.submit(
                lambda: asyncio.run(self._upload_resources(run, tracker))
            )
            upload.result()

    async def _download_resource(
        self, download_url: str, target_dir: str, session: aiohttp.ClientSession
    ) -> Any:
        download_url_path = urlparse(download_url).path
        relative_resource_part = _strip_resource_root_path(download_url_path)
        resource_dir, resource_file = os.path.split(relative_resource_part)
        download_dir = os.path.join(target_dir, resource_dir)
        Path(download_dir).mkdir(parents=True, exist_ok=True)
        download_path = os.path.join(download_dir, resource_file)
        async with session.get(
            download_url, timeout=aiohttp.ClientTimeout(total=None)
        ) as response:
            with open(download_path, mode="wb") as f:
                bytes_total = 0
                async for data in response.content.iter_chunked(8192):
                    bytes_total += f.write(data)
        self._log.info(
            f"Downloaded {download_url_path} to {download_path}, bytes total: {bytes_total}"
        )

    async def _download_resources(
        self, project_full_name: ProjectFullName, function_name: str, target_dir: str
    ) -> Any:
        resource_paths = self._client.data_catalog.get_resource_paths(
            project_full_name, function_name
        )
        download_root = os.getcwd() if target_dir == "" else target_dir
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            download_tasks = [
                self._download_resource(download_url, download_root, session)
                for download_url in resource_paths
            ]
            await asyncio.gather(*download_tasks)

    def wait_resource_download(
        self,
        project_full_name: ProjectFullName,
        function_name: str,
        target_dir: str = "",
    ) -> None:
        """
        Download resource files for the function.
        """
        # run in a separate thread to make sure no other loop is running
        with ThreadPoolExecutor(max_workers=1) as executor:
            download = executor.submit(
                lambda: asyncio.run(
                    self._download_resources(
                        project_full_name, function_name, target_dir
                    )
                )
            )
            download.result()



from datetime import timedelta
from enum import Enum
from typing import Any, List

import humanize  # type: ignore
from rich.console import Group, RenderableType
from rich.markup import escape
from rich.progress import ProgressColumn, Task
from rich.progress_bar import ProgressBar
from rich.table import Table
from rich.text import Text

from layer.contracts.tracker import (
    AssetTracker,
    AssetTrackerStatus,
    DatasetTransferState,
    ResourceTransferState,
)


class ProgressStyle(str, Enum):
    NONE = ""

    GREY = "rgb(161,161,169)"
    GREEN = "rgb(21,127,61)"
    ORANGE = "rgb(251,147,60)"
    BLUE = "rgb(37,99,234)"
    GRAY = "rgb(155,155,159)"
    BLACK = "rgb(0,0,0)"
    DEFAULT = "default"

    LINK = f"underline {GREY}"
    PENDING = GREY
    DONE = GREEN
    ERROR = ORANGE


class AssetColumn(ProgressColumn):
    _status_style_map = {
        AssetTrackerStatus.PENDING: ProgressStyle.NONE,
        AssetTrackerStatus.SAVING: ProgressStyle.NONE,
        AssetTrackerStatus.BUILDING: ProgressStyle.NONE,
        AssetTrackerStatus.TRAINING: ProgressStyle.NONE,
        AssetTrackerStatus.DONE: ProgressStyle.DONE,
        AssetTrackerStatus.ERROR: ProgressStyle.ERROR,
        AssetTrackerStatus.ASSERTING: ProgressStyle.NONE,
        AssetTrackerStatus.RESOURCE_UPLOADING: ProgressStyle.NONE,
        AssetTrackerStatus.RESULT_UPLOADING: ProgressStyle.NONE,
        AssetTrackerStatus.ASSET_DOWNLOADING: ProgressStyle.BLUE,
        AssetTrackerStatus.ASSET_FROM_CACHE: ProgressStyle.BLUE,
        AssetTrackerStatus.ASSET_LOADED: ProgressStyle.DONE,
    }

    def render(self, task: Task) -> RenderableType:
        asset = self._get_asset(task)
        description_text = self._render_description(task)
        stats_text = self._render_stats(task)

        table = Table.grid(padding=(0, 1, 0, 1), pad_edge=True, expand=False)
        table.add_column(width=20, min_width=20, max_width=20)  # name
        table.add_column(width=10, min_width=10, max_width=10)  # bar
        table.add_column(width=len(description_text))  # task description
        table.add_column()  # stats
        table.add_row(
            asset.name,
            self._render_progress_bar(task),
            description_text,
            stats_text,
        )
        renderables: List[RenderableType] = [table]
        if asset.base_url:
            table = Table.grid(padding=(0, 1, 0, 1), pad_edge=True)
            table.add_column()
            table.add_row(self._render_url(asset))
            renderables.append(table)
        if asset.error_reason:
            table = Table.grid(padding=(0, 1, 0, 1), pad_edge=True)
            table.add_column(overflow="fold")
            table.add_row(Text.from_markup(f"[red]{escape(asset.error_reason)}[/red]"))
            table.add_row(Text("Aborting...", style="bold"))
            renderables.append(table)

        return Group(*renderables)

    @staticmethod
    def _render_state(state: ResourceTransferState, show_num_files: bool = True) -> str:
        render_parts = []
        if show_num_files:
            render_parts.append(
                f"{state.transferred_num_files}/{state.total_num_files} files"
            )
        at_least_a_gig = state.total_resource_size_bytes > 1_000_000_000
        size_format = "%.1f" if at_least_a_gig else "%.f"
        transferred_bytes = humanize.naturalsize(
            state.transferred_resource_size_bytes, format=size_format
        )
        total_resources_bytes = humanize.naturalsize(
            state.total_resource_size_bytes, format=size_format
        )
        render_parts.append(f"{transferred_bytes}/{total_resources_bytes}")
        if state.transferred_num_files < state.total_num_files:
            render_parts.append(
                f"{humanize.naturalsize(state.get_bandwidth_in_previous_seconds(), format='%.1f')}/s"
            )
        return " | ".join(render_parts)

    @staticmethod
    def _render_dataset_state(state: DatasetTransferState) -> str:
        return f"{state.transferred_num_rows}/{state.total_num_rows} rows"

    def _render_progress_bar(self, task: Task) -> RenderableType:
        # Set task.completed = min(..., int(task.total -  1)) to prevent task timer from stopping
        asset = self._get_asset(task)
        style_map = {
            AssetTrackerStatus.PENDING: (
                True,
                ProgressStyle.DEFAULT,
                ProgressStyle.DONE,
            ),
            AssetTrackerStatus.SAVING: (
                True,
                ProgressStyle.DEFAULT,
                ProgressStyle.DONE,
            ),
            AssetTrackerStatus.BUILDING: (
                False,
                ProgressStyle.DEFAULT,
                ProgressStyle.BLACK,
            ),
            AssetTrackerStatus.TRAINING: (
                False,
                ProgressStyle.DEFAULT,
                ProgressStyle.BLACK,
            ),
            AssetTrackerStatus.DONE: (False, ProgressStyle.DONE, ProgressStyle.DONE),
            AssetTrackerStatus.ERROR: (
                False,
                ProgressStyle.ERROR,
                ProgressStyle.ERROR,
            ),
            AssetTrackerStatus.ASSERTING: (
                True,
                ProgressStyle.DEFAULT,
                ProgressStyle.DONE,
            ),
            AssetTrackerStatus.RESOURCE_UPLOADING: (
                False,
                ProgressStyle.DEFAULT,
                ProgressStyle.BLUE,
            ),
            AssetTrackerStatus.RESULT_UPLOADING: (
                False,
                ProgressStyle.DEFAULT,
                ProgressStyle.BLUE,
            ),
            AssetTrackerStatus.ASSET_DOWNLOADING: (
                False,
                ProgressStyle.DEFAULT,
                ProgressStyle.BLUE,
            ),
            AssetTrackerStatus.ASSET_FROM_CACHE: (
                True,
                ProgressStyle.DEFAULT,
                ProgressStyle.DONE,
            ),
            AssetTrackerStatus.ASSET_LOADED: (
                False,
                ProgressStyle.DEFAULT,
                ProgressStyle.DONE,
            ),
        }
        pulse, style, completed_style = style_map.get(
            asset.status, (False, ProgressStyle.DEFAULT, ProgressStyle.DONE)
        )
        if asset.status == AssetTrackerStatus.ASSET_DOWNLOADING and isinstance(
            asset.asset_download_transfer_state, DatasetTransferState
        ):
            pulse = True  # Pulsing bar as we currently cannot closely track the downloading of a dataset

        assert task.total is not None
        if (
            asset.status == AssetTrackerStatus.TRAINING
            or asset.status == AssetTrackerStatus.BUILDING
        ):
            fraction = round(task.total * 0.05)
            task.completed = min(
                (task.completed + fraction) % task.total, task.total - 1
            )
        elif asset.status == AssetTrackerStatus.RESOURCE_UPLOADING:
            assert asset.resource_transfer_state
            state = asset.resource_transfer_state
            completed = (
                (
                    state.transferred_resource_size_bytes
                    / state.total_resource_size_bytes
                )
                if state.total_resource_size_bytes > 0
                else 0
            )
            task.completed = min(int(task.total * completed), int(task.total - 1))
        elif asset.status == AssetTrackerStatus.ASSET_LOADED:
            task.completed = task.total
        elif (
            asset.status == AssetTrackerStatus.RESULT_UPLOADING
            and asset.dataset_transfer_state
        ):
            dataset_state = asset.dataset_transfer_state
            completed = (
                (dataset_state.transferred_num_rows / dataset_state.total_num_rows)
                if dataset_state.total_num_rows > 0
                else 0
            )
            task.completed = min(int(task.total * completed), int(task.total - 1))
        elif (
            asset.status == AssetTrackerStatus.RESULT_UPLOADING
        ) and asset.model_transfer_state:
            model_state = asset.model_transfer_state
            completed = (
                (
                    model_state.transferred_resource_size_bytes
                    / model_state.total_resource_size_bytes
                )
                if model_state.total_resource_size_bytes > 0
                else 0
            )
            task.completed = min(int(task.total * completed), int(task.total - 1))
        elif (
            (asset.status == AssetTrackerStatus.ASSET_DOWNLOADING)
            and asset.asset_download_transfer_state
            and isinstance(asset.asset_download_transfer_state, ResourceTransferState)
        ):
            model_state = asset.asset_download_transfer_state
            completed = (
                (
                    model_state.transferred_resource_size_bytes
                    / model_state.total_resource_size_bytes
                )
                if model_state.total_resource_size_bytes > 0
                else 0
            )
            task.completed = min(int(task.total * completed), int(task.total - 1))

        return ProgressBar(
            total=max(0.0, task.total),
            completed=max(0.0, task.completed),
            animation_time=task.get_time(),
            style=style,
            pulse=pulse,
            complete_style=completed_style,
            finished_style=completed_style,
            pulse_style=ProgressStyle.PENDING,
        )

    @staticmethod
    def _get_asset(task: Task) -> AssetTracker:
        return task.fields["asset"]

    @staticmethod
    def _get_elapsed_time_s(task: Task) -> int:
        elapsed_time = task.finished_time if task.finished else task.elapsed
        return int(elapsed_time) if elapsed_time else 0

    def _compute_time_string(self, task: Task) -> str:
        asset = self._get_asset(task)
        delta: Any = timedelta(seconds=self._get_elapsed_time_s(task))
        if asset.status == AssetTrackerStatus.RESOURCE_UPLOADING:
            assert asset.resource_transfer_state
            delta = timedelta(seconds=asset.resource_transfer_state.get_eta_seconds())
        elif (
            asset.status == AssetTrackerStatus.RESULT_UPLOADING
            and asset.dataset_transfer_state
        ):
            delta = timedelta(seconds=asset.dataset_transfer_state.get_eta_seconds())
        elif (
            asset.status == AssetTrackerStatus.RESULT_UPLOADING
            and asset.model_transfer_state
        ):
            delta = timedelta(seconds=asset.model_transfer_state.get_eta_seconds())
        elif asset.status == AssetTrackerStatus.ASSET_DOWNLOADING:
            assert asset.asset_download_transfer_state
            if isinstance(asset.asset_download_transfer_state, ResourceTransferState):
                delta = timedelta(
                    seconds=asset.asset_download_transfer_state.get_eta_seconds()
                )
            else:
                delta = "-:--:--"
        elif (
            asset.status == AssetTrackerStatus.PENDING
            or asset.status == AssetTrackerStatus.ASSET_FROM_CACHE
        ):
            delta = "-:--:--"
        elif (
            asset.status == AssetTrackerStatus.TRAINING
            or asset.status == AssetTrackerStatus.BUILDING
        ):
            delta = timedelta(seconds=self._get_elapsed_time_s(task))
        return str(delta)

    def _render_stats(self, task: Task) -> RenderableType:
        asset = self._get_asset(task)
        rendered_time: str = self._compute_time_string(task)
        rendered_state = None

        if (
            asset.resource_transfer_state
            and asset.status == AssetTrackerStatus.RESOURCE_UPLOADING
        ):
            rendered_state = self._render_state(asset.resource_transfer_state)
        elif (
            asset.dataset_transfer_state
            and asset.status == AssetTrackerStatus.RESULT_UPLOADING
        ):
            rendered_state = self._render_dataset_state(asset.dataset_transfer_state)
        elif (
            asset.model_transfer_state
            and asset.status == AssetTrackerStatus.RESULT_UPLOADING
        ):
            rendered_state = self._render_state(asset.model_transfer_state, False)
        elif asset.status == AssetTrackerStatus.ASSET_DOWNLOADING:
            assert asset.asset_download_transfer_state
            if isinstance(asset.asset_download_transfer_state, ResourceTransferState):
                rendered_state = self._render_state(
                    asset.asset_download_transfer_state, False
                )

        color = ProgressStyle.GRAY
        stats = (
            [rendered_time]
            if rendered_state is None
            else [rendered_state, rendered_time]
        )
        stats = (" | ".join(stats)).split(" | ")
        stats = [f"[{color}]{stat}[/{color}]" for stat in stats]
        text = ", ".join(stats)
        return Text.from_markup(f"[{text}]", style="default", justify="center")

    def _render_description(self, task: Task) -> Text:
        asset = self._get_asset(task)
        if (
            asset.status == AssetTrackerStatus.RESOURCE_UPLOADING
            or asset.status == AssetTrackerStatus.RESULT_UPLOADING
        ):
            text = "uploading"
        elif asset.status == AssetTrackerStatus.ASSET_DOWNLOADING:
            text = "downloading"
        elif asset.status == AssetTrackerStatus.ASSET_FROM_CACHE:
            text = "from cache"
        elif asset.status == AssetTrackerStatus.ASSET_LOADED:
            text = "loaded"
        else:
            text = task.description
        if (
            asset.status == AssetTrackerStatus.RESOURCE_UPLOADING
            or asset.status == AssetTrackerStatus.RESULT_UPLOADING
            or asset.status == AssetTrackerStatus.ASSET_DOWNLOADING
            or asset.status == AssetTrackerStatus.ASSET_FROM_CACHE
        ):
            style = ProgressStyle.BLACK
        else:
            style = self._status_style_map[asset.status]
        return Text(
            text.upper(),
            overflow="fold",
            style=style,
            justify="center",
        )

    @staticmethod
    def _render_url(asset: AssetTracker) -> RenderableType:
        link = (
            f"{asset.base_url}?v={asset.version}.{asset.build_idx}"
            if (asset.version and asset.build_idx)
            else asset.base_url
        )
        return Group(
            *[
                Text.from_markup(
                    f"↳ [link={str(link)}]{str(link)}[/link]",
                    style=ProgressStyle.LINK,
                    overflow="fold",
                    justify="default",
                )
            ]
        )



from .config import (  # noqa
    DEFAULT_FUNC_PATH,
    DEFAULT_LAYER_PATH,
    DEFAULT_PATH,
    DEFAULT_URL,
    AccountServiceConfig,
    AuthConfig,
    ClientConfig,
    Config,
    ConfigRecord,
    ConfigStore,
    Credentials,
    DataCatalogConfig,
    FlowManagerServiceConfig,
    LogsConfig,
    ModelCatalogConfig,
    ModelTrainingConfig,
    ProjectServiceConfig,
    S3Config,
    UserLogsServiceConfig,
)
from .config_client import ConfigClient  # noqa
from .config_manager import ConfigManager  # noqa



from dataclasses import dataclass
from typing import Any, Callable, List


@dataclass(frozen=True)
class Assertion:
    name: str
    values: List[Any]
    function: Callable[..., Any]

    def __str__(self) -> str:
        values_str = []
        for value in self.values:
            if callable(value):
                values_str.append(value.__name__)
            else:
                values_str.append(str(value))
        return f"{self.name}({', '.join(values_str)})"



from dataclasses import replace
from pathlib import Path
from typing import Optional, Type

from yarl import URL

from layer.auth import (
    CodeClient,
    CredentialsClient,
    HeadlessCodeClient,
    WebBrowserCodeClient,
)
from layer.config import DEFAULT_PATH, DEFAULT_URL, Config, ConfigStore, Credentials
from layer.exceptions.exceptions import (
    InvalidConfigurationError,
    MissingConfigurationError,
    UserAccessTokenExpiredError,
    UserConfigurationError,
    UserNotLoggedInException,
    UserWithoutAccountError,
)

from .config_client import ConfigClient
from .guest_login_client import get_guest_auth_token


class ConfigManager:
    def __init__(self, path: Optional[Path] = None) -> None:
        if path is None:
            path = self._get_default_path()
        assert path
        self._store = ConfigStore(path)

    @staticmethod
    def _get_default_path() -> Path:
        return DEFAULT_PATH

    def load(self) -> Config:
        return self._store.load()

    async def refresh(self, *, allow_guest: bool = False) -> Config:
        from aiohttp import ClientSession

        try:
            config = self._store.load()
        except MissingConfigurationError:
            if allow_guest:
                config = await self.login_as_guest(DEFAULT_URL)
            else:
                await self._logout()
                raise UserNotLoggedInException()

        except InvalidConfigurationError as cex:
            await self._logout()
            raise UserConfigurationError(cex.path)

        if not config.auth.is_enabled or not config.credentials.is_access_token_expired:
            return config

        async with ClientSession() as client:
            creds_client = CredentialsClient(
                client=client,
                url=config.auth.token_url,
                client_id=config.auth.client_id,
            )
            creds = await creds_client.refresh(config.credentials)

        config = config.with_credentials(creds)
        self._store.save(config)
        return config

    async def _login(
        self, url: URL, code_client_factory: Type[CodeClient] = HeadlessCodeClient
    ) -> Config:
        from aiohttp import ClientSession

        await self._logout()

        async with ClientSession() as client:
            config_client = ConfigClient(client=client, url=url)
            config = await config_client.get_config()

            code_client = code_client_factory(  # pytype: disable=not-instantiable
                config=config.auth
            )
            code = await code_client.request()

            creds_client = CredentialsClient(
                client=client,
                url=config.auth.token_url,
                client_id=config.auth.client_id,
            )
            creds = await creds_client.request(code)

            if creds.is_authenticated_outside_organization:
                raise UserWithoutAccountError(config.url)

        config = config.with_credentials(creds)
        self._store.save(config)
        print(f"Successfully logged into {url}")
        return config

    async def _logout(self) -> None:
        self._store.delete()

    async def login(self, url: URL) -> Config:
        return await self._login(url, code_client_factory=WebBrowserCodeClient)

    async def login_as_guest(self, url: URL) -> Config:
        _url = url.with_host(f"grpc.{url.host}")
        grpc_url = f"{_url.host}:{_url.port}"

        token = get_guest_auth_token(str(grpc_url))

        return await self.login_with_access_token(url, token, config_is_guest=True)

    async def login_headless(self, url: URL) -> Config:
        return await self._login(url, code_client_factory=HeadlessCodeClient)

    async def login_with_access_token(
        self, url: URL, access_token: str, config_is_guest: bool = False
    ) -> Config:
        from aiohttp import ClientSession

        await self._logout()

        async with ClientSession() as client:
            config_client = ConfigClient(client=client, url=url)
            config = await config_client.get_config()

        creds = Credentials(access_token=access_token, refresh_token="")
        if config.auth.is_enabled:
            if creds.is_access_token_expired:
                raise UserAccessTokenExpiredError()

        config = config.with_credentials(creds)
        config = replace(config, is_guest=config_is_guest)
        self._store.save(config)
        if config_is_guest:
            print(f"Successfully logged into {url} as guest")
        else:
            print(f"Successfully logged into {url}")
        return config

    async def login_with_api_key(self, url: URL, api_key: str) -> Config:
        await self._logout()

        from aiohttp import ClientSession

        async with ClientSession() as client:
            config_client = ConfigClient(client=client, url=url)
            config = await config_client.get_config()

        creds = Credentials(access_token="", refresh_token=api_key)
        config = config.with_credentials(creds)

        self._store.save(config)
        await self.refresh()
        print(f"Successfully logged into {url}")
        return config

    async def login_with_config(self, config: Config) -> Config:
        await self._logout()
        self._store.save(config)
        print(f"Successfully logged into {config.url}")
        return config

    async def logout(self) -> None:
        msg = "Logged out."
        try:
            self._store.load()
        except MissingConfigurationError:
            msg = "No active sessions."
        except InvalidConfigurationError:
            pass

        await self._logout()
        print(msg)



from dataclasses import dataclass
from enum import Enum, unique
from typing import Dict


@unique
class LoggedDataType(Enum):
    INVALID = 0
    TEXT = 1
    TABLE = 2
    BLOB = 3
    NUMBER = 4
    BOOLEAN = 5
    IMAGE = 6
    VIDEO = 7
    MARKDOWN = 8


@dataclass(frozen=True)
class LoggedData:
    logged_data_type: LoggedDataType
    tag: str
    data: str
    epoched_data: Dict[int, str]


@dataclass(frozen=True)
class ModelMetricPoint:
    epoch: int
    value: float


@dataclass(frozen=True)
class Markdown:
    """
    Encapsulates a markdown text to be displayed correctly in Layer UI as part of logging.

    .. code-block:: python

        layer.log(layer.Markdown("## Hello world!"))

    """

    data: str



import sys
import tempfile
import warnings
from contextlib import contextmanager
from logging import Logger
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Optional, Tuple

from layerapi.api.entity.model_pb2 import Model as PBModel
from layerapi.api.entity.model_train_pb2 import ModelTrain as PBModelTrain
from layerapi.api.entity.model_train_status_pb2 import (
    ModelTrainStatus as PBModelTrainStatus,
)
from layerapi.api.entity.model_version_pb2 import ModelVersion
from layerapi.api.entity.source_code_environment_pb2 import SourceCodeEnvironment
from layerapi.api.ids_pb2 import ModelTrainId, ModelVersionId
from layerapi.api.service.modelcatalog.model_catalog_api_pb2 import (
    CompleteModelTrainRequest,
    CreateModelTrainFromVersionIdRequest,
    CreateModelTrainRequest,
    CreateModelVersionRequest,
    CreateModelVersionResponse,
    GetModelByPathRequest,
    GetModelByPathResponse,
    GetModelTrainRequest,
    GetModelTrainResponse,
    GetModelTrainStorageConfigurationRequest,
    GetModelVersionRequest,
    GetModelVersionResponse,
    LoadModelTrainDataByPathRequest,
    StartModelTrainRequest,
    StoreTrainingMetadataRequest,
    UpdateModelTrainStatusRequest,
)
from layerapi.api.service.modelcatalog.model_catalog_api_pb2_grpc import (
    ModelCatalogAPIStub,
)
from layerapi.api.value.dependency_pb2 import DependencyFile
from layerapi.api.value.s3_path_pb2 import S3Path
from layerapi.api.value.sha256_pb2 import Sha256
from layerapi.api.value.source_code_pb2 import RemoteFileLocation, SourceCode

from layer.cache.cache import Cache
from layer.config import ClientConfig
from layer.contracts.models import Model, ModelObject, TrainStorageConfiguration
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.runs import FunctionDefinition
from layer.contracts.tracker import ResourceTransferState
from layer.exceptions.exceptions import LayerClientException
from layer.flavors.base import ModelRuntimeObjects
from layer.flavors.utils import get_flavor_for_proto
from layer.tracker.progress_tracker import RunProgressTracker
from layer.utils.grpc import create_grpc_channel
from layer.utils.s3 import S3Util


if TYPE_CHECKING:
    from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PBModelFlavor


class ModelCatalogClient:
    _service: ModelCatalogAPIStub

    def __init__(
        self, config: ClientConfig, logger: Logger, cache_dir: Optional[Path] = None
    ):
        self._config = config.model_catalog
        self._s3_endpoint_url = config.s3.endpoint_url
        self._logger = logger
        self._access_token = config.access_token
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

        self._cache = Cache(cache_dir).initialise()

    @contextmanager
    def init(self) -> Iterator["ModelCatalogClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = ModelCatalogAPIStub(channel=channel)
            yield self

    def create_model_version(
        self,
        project_full_name: ProjectFullName,
        model: FunctionDefinition,
        is_local: bool,
    ) -> CreateModelVersionResponse:
        """
        Given a model metadata it makes a request to the backend
        and creates a corresponding entity.
        :param project_full_name: the project full name of the model
        :param model: the structured of the parsed entity
        :return: the created model version entity
        """
        model_path = f"{project_full_name.path}/models/{model.asset_name}"
        self._logger.debug(
            f"Creating model version for the following model: {model_path}"
        )
        response = self._service.CreateModelVersion(
            CreateModelVersionRequest(
                model_path=model_path,
                description=model.description,
                training_files_hash=Sha256(value=model.source_code_digest.hexdigest()),
                should_create_initial_train=True,
                fabric=model.get_fabric(is_local=is_local),
            ),
        )
        self._logger.debug(f"CreateModelVersionResponse: {str(response)}")
        return response

    def store_training_metadata(
        self,
        model: FunctionDefinition,
        s3_path: S3Path,
        version: ModelVersion,
    ) -> None:
        language_version = _language_version()
        request: StoreTrainingMetadataRequest = StoreTrainingMetadataRequest(
            model_version_id=version.id,
            name=model.asset_name,
            description=model.description,
            source_code_env=SourceCodeEnvironment(
                source_code=SourceCode(
                    remote_file_location=RemoteFileLocation(
                        name=model.entrypoint,
                        location=f"s3://{s3_path.bucket}/{s3_path.key}{model.asset_name}.tgz",
                    ),
                    language=SourceCode.Language.LANGUAGE_PYTHON,
                    language_version=SourceCode.LanguageVersion(
                        major=int(language_version[0]),
                        minor=int(language_version[1]),
                        micro=int(language_version[2]),
                    ),
                ),
                dependency_file=DependencyFile(
                    name=model.environment,
                    location=model.environment,
                ),
            ),
            entrypoint=model.entrypoint,
        )
        self._logger.debug(f"StoreTrainingMetadataRequest request: {str(request)}")
        response = self._service.StoreTrainingMetadata(request)
        self._logger.debug(f"StoreTrainingMetadata response: {str(response)}")

    def load_model_by_path(self, path: str) -> Model:
        load_response = self._service.LoadModelTrainDataByPath(
            LoadModelTrainDataByPathRequest(path=path),
        )

        flavor = get_flavor_for_proto(load_response.flavor)
        if flavor is None:
            raise LayerClientException(
                f"Unexpected model flavor {type(load_response.flavor)}"
            )
        return Model(
            asset_path=path,
            id=load_response.id.value,
            flavor=flavor,
            storage_config=TrainStorageConfiguration(
                train_id=load_response.id,
                s3_path=load_response.s3_path,
                credentials=load_response.credentials,
            ),
        )

    def create_model_train_from_version_id(
        self,
        version_id: ModelVersionId,
    ) -> ModelTrainId:
        response = self._service.CreateModelTrainFromVersionId(
            CreateModelTrainFromVersionIdRequest(
                model_version_id=version_id,
            ),
        )
        return response.id

    def load_model_runtime_objects(
        self,
        model: Model,
        state: ResourceTransferState,
        no_cache: bool = False,
    ) -> ModelRuntimeObjects:
        """
        Loads a model artifact from the model catalog

        :param model: the model
        :return: the model artifact
        """
        self._logger.debug(f"Loading model {model.path}")
        try:
            model_cache_dir = self._cache.get_path_entry(str(model.id))
            if no_cache or model_cache_dir is None:
                with tempfile.TemporaryDirectory() as tmp:
                    local_path = Path(tmp) / "model"

                    S3Util.download_dir(
                        local_dir=local_path,
                        credentials=model.storage_config.credentials,
                        s3_path=model.storage_config.s3_path,
                        endpoint_url=self._s3_endpoint_url,
                        state=state,
                    )
                    if no_cache:
                        return self._load_model_runtime_objects(model, local_path)
                    model_cache_dir = self._cache.put_path_entry(
                        str(model.id), local_path
                    )

            assert model_cache_dir is not None
            return self._load_model_runtime_objects(model, model_cache_dir)
        except Exception as ex:
            raise LayerClientException(f"Error while loading model, {ex}")

    def _load_model_runtime_objects(
        self, model: Model, model_dir: Path
    ) -> ModelRuntimeObjects:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            return model.flavor.load_model_from_directory(model_dir)

    def save_model_object(
        self,
        model: Model,
        model_object: ModelObject,
        tracker: RunProgressTracker,
    ) -> ModelObject:
        self._logger.debug(f"Storing given model {model_object} for {model.path}")
        try:
            with tempfile.TemporaryDirectory() as tmp:
                local_path = Path(tmp) / "model"
                model.flavor.save_model_to_directory(model_object, local_path)
                state = ResourceTransferState()
                tracker.mark_model_saving_result(model.name, state)
                S3Util.upload_dir(
                    local_dir=local_path,
                    credentials=model.storage_config.credentials,
                    s3_path=model.storage_config.s3_path,
                    endpoint_url=self._s3_endpoint_url,
                    state=state,
                )
        except Exception as ex:
            raise LayerClientException(f"Error while storing model, {ex}")
        self._logger.debug(f"User model {model.path} saved successfully")

        return model_object

    def get_model_train_storage_configuration(
        self,
        train_id: ModelTrainId,
    ) -> TrainStorageConfiguration:
        response = self._service.GetModelTrainStorageConfiguration(
            GetModelTrainStorageConfigurationRequest(
                model_train_id=train_id,
            ),
        )
        return TrainStorageConfiguration(
            train_id=train_id,
            s3_path=response.s3_path,
            credentials=response.credentials,
        )

    def start_model_train(
        self,
        train_id: ModelTrainId,
    ) -> TrainStorageConfiguration:
        response = self._service.StartModelTrain(
            StartModelTrainRequest(
                model_train_id=train_id,
            ),
        )
        return TrainStorageConfiguration(
            train_id=train_id,
            s3_path=response.s3_path,
            credentials=response.credentials,
        )

    def create_model_train(
        self,
        name: str,
        project_full_name: ProjectFullName,
        version: Optional[str],
    ) -> ModelTrainId:
        response = self._service.CreateModelTrain(
            CreateModelTrainRequest(
                model_name=name,
                model_version="" if version is None else version,
                project_full_name=project_full_name.path,
            ),
        )
        return response.id

    def complete_model_train(
        self,
        train_id: ModelTrainId,
        flavor: Optional["PBModelFlavor.ValueType"],
    ) -> None:
        self._service.CompleteModelTrain(
            CompleteModelTrainRequest(id=train_id, flavor=flavor),
        )

    def get_model_by_path(self, model_path: str) -> PBModel:
        response: GetModelByPathResponse = self._service.GetModelByPath(
            GetModelByPathRequest(
                path=model_path,
            )
        )
        return response.model

    def get_model_train(self, train_id: ModelTrainId) -> PBModelTrain:
        response: GetModelTrainResponse = self._service.GetModelTrain(
            GetModelTrainRequest(
                model_train_id=train_id,
            ),
        )
        return response.model_train

    def get_model_version(self, version_id: ModelVersionId) -> ModelVersion:
        response: GetModelVersionResponse = self._service.GetModelVersion(
            GetModelVersionRequest(
                model_version_id=version_id,
            ),
        )
        return response.version

    def update_model_train_status(
        self, train_id: ModelTrainId, train_status: PBModelTrainStatus
    ) -> None:
        self._service.UpdateModelTrainStatus(
            UpdateModelTrainStatusRequest(
                model_train_id=train_id, train_status=train_status
            )
        )


def _language_version() -> Tuple[int, int, int]:
    return sys.version_info.major, sys.version_info.minor, sys.version_info.micro



import uuid
from contextlib import contextmanager
from logging import Logger
from typing import TYPE_CHECKING, Iterator, List, Tuple

from layerapi.api.entity.user_log_line_pb2 import UserLogLine
from layerapi.api.ids_pb2 import RunId
from layerapi.api.service.user_logs.user_logs_api_pb2 import (
    GetPipelineRunLogsRequest,
    GetPipelineRunLogsResponse,
)
from layerapi.api.service.user_logs.user_logs_api_pb2_grpc import UserLogsAPIStub

from layer.config import ClientConfig
from layer.utils.grpc import create_grpc_channel


if TYPE_CHECKING:
    pass


class UserLogsClient:
    _service: UserLogsAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
    ):
        self._config = config.user_logs
        self._logger = logger
        self._access_token = config.access_token
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

    @contextmanager
    def init(self) -> Iterator["UserLogsClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
            options=[
                (
                    "grpc.max_receive_message_length",
                    self._config.max_receive_message_length,
                )
            ],
        ) as channel:
            self._service = UserLogsAPIStub(channel=channel)
            yield self

    def get_pipeline_run_logs(
        self, run_id: uuid.UUID, continuation_token: str
    ) -> Tuple[List[UserLogLine], str]:
        response: GetPipelineRunLogsResponse = self._service.GetPipelineRunLogs(
            request=GetPipelineRunLogsRequest(
                run_id=RunId(value=str(run_id)), continuation_token=continuation_token
            )
        )
        return list(response.log_lines), response.continuation_token



from pathlib import Path
from traceback import FrameSummary
from typing import Any, List, Optional, Type

from layerapi.api.ids_pb2 import RunId
from yarl import URL

from layer.contracts.assertions import Assertion

from .status_report import AssertionFailureStatusReport, ExecutionStatusReport


RICH_ERROR_COLOR = "rgb(251,147,60)"


class RuntimeMemoryException(Exception):
    def __init__(self, err_details: str):
        super().__init__(f"Memory error: {err_details}")


class ConfigError(Exception):
    pass


class MissingConfigurationError(ConfigError):
    def __init__(self, path: Path):
        self._path: Path = path
        super().__init__(f"Missing configuration file {path.resolve()}")

    @property
    def path(self) -> Path:
        return self._path


class InvalidConfigurationError(ConfigError):
    def __init__(self, path: Path):
        self._path: Path = path
        super().__init__(f"Invalid configuration file {path.resolve()}")

    @property
    def path(self) -> Path:
        return self._path


class AuthException(Exception):
    pass


class LayerClientException(Exception):
    pass


class LayerClientTimeoutException(LayerClientException):
    pass


class LayerClientResourceNotFoundException(LayerClientException):
    pass


class LayerClientResourceAlreadyExistsException(LayerClientException):
    pass


class LayerResourceExhaustedException(LayerClientException):
    pass


class LayerClientServiceUnavailableException(LayerClientException):
    pass


class ProjectException(Exception):
    pass


class ProjectBaseException(Exception):
    def __init__(self, error_msg: str = "", suggestion: str = ""):
        self._error_msg = error_msg
        self._suggestion = suggestion
        super().__init__(self._format_message())

    @property
    def error_msg(self) -> str:
        return self._error_msg

    @property
    def error_msg_rich(self) -> str:
        return self.error_msg

    @property
    def suggestion(self) -> str:
        return self._suggestion

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def _format_message(self) -> str:
        return f"Error: {self._error_msg},\nSuggestion: {self._suggestion}"


class ProjectExecutionException(ProjectBaseException):
    def __init__(self, run_id: RunId, err_msg: str, suggestion: str):
        super().__init__(err_msg, suggestion)
        self._run_id = run_id

    def run_id(self) -> RunId:
        return self._run_id


class ProjectModelExecutionException(ProjectExecutionException):
    def __init__(self, run_id: RunId, train_id: str, report: ExecutionStatusReport):
        self._report = report
        self._train_id = train_id
        super().__init__(
            run_id,
            self.__format_message(),
            "Add `debug=True` parameter to your `layer.run()` to see server logs",
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f"Train [bold {RICH_ERROR_COLOR}]{self._train_id}[/bold {RICH_ERROR_COLOR}] failed with "
            + f'"{self._report.message}"'
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return f"Train {self._train_id} failed with " + f'"{self._report.message}"'

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectHPTExecutionException(ProjectExecutionException):
    def __init__(self, run_id: RunId, tuning_id: str, report: ExecutionStatusReport):
        self._tuning_id = tuning_id
        self._report = report
        super().__init__(
            run_id,
            self.__format_message(),
            "Add `debug=True` parameter to your `layer.run()` to see server logs",
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f"HPT [bold {RICH_ERROR_COLOR}]{self._tuning_id}[/bold {RICH_ERROR_COLOR}] failed with "
            + f'"{self._report.message}"'
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return f"HPT {self._tuning_id} failed with " + f'"{self._report.message}"'

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectDatasetBuildExecutionException(ProjectExecutionException):
    def __init__(self, run_id: RunId, dataset_id: str, report: ExecutionStatusReport):
        self._dataset_id = dataset_id
        self._report = report
        super().__init__(
            run_id,
            self.__format_message(),
            "Add `debug=True` parameter to your `layer.run()` to see server logs",
        )

    @property
    def message(self) -> str:
        return self._report.message

    @property
    def error_msg_rich(self) -> str:
        return (
            f'Dataset "[bold {RICH_ERROR_COLOR}]{self._dataset_id}[/bold {RICH_ERROR_COLOR}]" build failed with '
            + f'"{self._report.message}"'
        )

    @property
    def suggestion_rich(self) -> str:
        return self.suggestion

    def __format_message(self) -> str:
        return (
            f'Dataset "{self._dataset_id}" build failed with '
            + f'"{self._report.message}"'
        )

    @staticmethod
    def __format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'


class ProjectInitializationException(ProjectBaseException):
    pass


class ProjectDependencyNotFoundException(ProjectInitializationException):
    pass


class LayerServiceUnavailableExceptionDuringInitialization(
    ProjectInitializationException
):
    def __init__(self, message: str):
        super().__init__(
            f"Layer service unavailable. {message}",
            "Check your internet connection.",
        )


class LayerServiceUnavailableExceptionDuringExecution(ProjectExecutionException):
    def __init__(self, run_id: RunId, message: str):
        super().__init__(
            run_id,
            f"Layer service unavailable. {message}",
            "Check your internet connection.",
        )


class UserNotLoggedInException(ProjectInitializationException):
    def __init__(self) -> None:
        super().__init__("User not logged in.", "Run `layer.login()`")


class UserConfigurationError(ProjectInitializationException):
    def __init__(self, path: Path) -> None:
        super().__init__(
            f"User configuration error in {path.resolve()}",
            "Please make sure that you have the latest version and try to login again.",
        )


class UserWithoutAccountError(ProjectInitializationException):
    def __init__(self, url: URL) -> None:
        super().__init__(
            "You don't have a Layer account yet",
            f"You can create your account and choose a username at {url}.",
        )


class UserAccessTokenExpiredError(ProjectInitializationException):
    def __init__(
        self,
    ) -> None:
        super().__init__("Your session has expired.", "Please, try to login again.")


class ProjectRunnerError(Exception):
    def __init__(self, message: str = "", run_id: Optional[RunId] = None):
        super().__init__(message)
        self._run_id = run_id

    def run_id(self) -> Optional[RunId]:
        return self._run_id


class ProjectRunTerminatedError(ProjectRunnerError):
    pass


class UnexpectedModelTypeException(Exception):
    def __init__(self, model_object_type: Type[Any]):
        super().__init__(
            f"train_model function returned an unsupported model object type '{model_object_type}'"
        )


class ProjectCircularDependenciesException(ProjectInitializationException):
    def __init__(self, stringified_paths: List[str]):
        super().__init__(
            self._err_msg(stringified_paths),
            "Please, check your project dependencies for errors",
        )
        self._paths: List[str] = stringified_paths

    @property
    def stringified_cycle_paths(self) -> List[str]:
        return self._paths

    @staticmethod
    def _err_msg(stringified_paths: List[str]) -> str:
        paths_as_str = "\n".join(stringified_paths)
        return f"Detected circular dependencies:\n{paths_as_str}"


class LayerFailedAssertionsException(Exception):
    def __init__(self, failed_assertions: List[Assertion]) -> None:
        super().__init__()
        self._failed_assertions = failed_assertions

    @property
    def failed_assertions(self) -> List[Assertion]:
        return self._failed_assertions

    def to_status_report(self) -> AssertionFailureStatusReport:
        return AssertionFailureStatusReport(failed_assertions=self.failed_assertions)



from abc import ABC
from typing import Any, Callable, Dict, List

import wrapt  # type: ignore

from layer.contracts.runs import ResourcePath
from layer.decorators.layer_wrapper import LayerFunctionWrapper


def resources(path: str, *paths: str) -> Callable[..., Any]:
    """
    Syncs arbitrary files between local and fabric environments.
    ``@resources`` decorator can be applied to the functions that are wrapped by the ``@dataset`` or ``@model`` decorators.

    :param path: File or directory path, could be relative or absolute. If path is a directory, all directories and files are uploaded recursively.
    :param paths: Additional paths.
    :return: Function object being decorated.

    .. code-block:: python

        import pandas as pd
        from layer
        from layer.decorators import dataset, resources

        @dataset("titanic")
        @resources("data/titanic.csv")
        def titanic():
            return pd.read_csv("data/titanic.csv")

    When you execute ``layer.run([titanic])``, the following happens:
    - Function ``titanic`` is submitted to Layer fabric, ``data/titanic.csv`` is uploaded to internal resource storage.
    - Before the execution, fabric runtime downloads the resource ``data/titanic.csv`` and makes it available to use within the function scope.

    If you have multiple resources, separate them with commas.

    .. code-block:: python

        @resources("path", "path1", "path2")

    **Note:** while absolute or relative paths are allowed as parameters to ``@resources``, they should always be relative when used within function body. For example:

    .. code-block:: python

        @dataset("titanic")
        @resources("/data/titanic.csv")  # it's ok to use absolute paths as they're resolve on local machine
        def titanic():
            return pd.read_csv("data/titanic.csv")  # while run inside function fabric, always use the relative paths!

    #### Limitations:
    - You can upload a maximum of 1000 files
    - Total upload size is limited to 5GB

    """

    # first arg is callable in case arg list is empty
    if not isinstance(path, str):
        raise ValueError("resource paths cannot be empty")

    @wrapt.decorator(proxy=_resources_wrapper(path, *paths))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _resources_wrapper(path: str, *paths: str) -> Any:
    class ResourcesFunctionWrapper(LayerFunctionWrapper, ABC):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any) -> None:
            super().__init__(wrapped, wrapper, enabled)

            _paths: List[ResourcePath] = []
            for _path in [path, *paths]:
                _paths.append(ResourcePath(path=_path))
            self.layer.set_resource_paths(_paths)

    return ResourcesFunctionWrapper



from typing import Union

from layer.contracts.datasets import Dataset
from layer.contracts.models import Model

from .cache import Cache


def is_cached(asset: Union[Dataset, Model]) -> bool:
    cache = Cache(cache_dir=None).initialise()
    cache_dir = cache.get_path_entry(str(asset.id))
    return cache_dir is not None


def clear_cache() -> None:
    """
    Clear all cached objects fetched by the Layer SDK on this machine.

    The Layer SDK locally stores all datasets and models by default on your computer.
    When you fetch a dataset with :func:``layer.get_dataset``, or load the model with ``layer.get_model``,
    the first call will fetch the artifact from remote storage,
    but subsequent calls will read it from the local disk.
    """
    Cache().clear()



import json
import re
from typing import Any, Dict, Optional

from layer.exceptions.exceptions import LayerClientException


def _try_parse_grpc_debug_context(message: str) -> Optional[Dict[str, Any]]:
    json_expr = re.search(r"\s({.*?})$", message)
    if json_expr:
        # NOTE: the replace call below is needed to mitigate a bug in gRPC on
        # Windows where backslashes in paths "file":"C:\\..." go unescaped
        json_context = json_expr.group(1).replace("\\", "\\\\")
        try:
            return json.loads(json_context)
        except ValueError:
            return None
    return None


def generate_client_error_from_grpc_error(
    err: Exception, internal_message: str
) -> LayerClientException:
    context = _try_parse_grpc_debug_context(str(err))
    if context:
        grpc_message = context.get("grpc_message")
        if grpc_message is not None:
            return LayerClientException(grpc_message)
    return LayerClientException(f"{internal_message}: {err}")



import json
import traceback
from abc import ABC
from os import path
from pathlib import Path
from traceback import FrameSummary
from typing import Any, List, Optional

from layer.contracts.assertions import Assertion


class ExecutionStatusReport(ABC):
    @property
    def message(self) -> str:
        return ""

    @property
    def cause(self) -> str:
        return ""


class GenericExecutionStatusReport(ExecutionStatusReport):
    def __init__(self, message: str):
        self._message = message

    @property
    def message(self) -> str:
        return self._message

    @property
    def cause(self) -> str:
        return ""

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, GenericExecutionStatusReport):
            return False
        return self.message == other.message


class PythonExecutionStatusReport(ExecutionStatusReport):
    def __init__(
        self,
        message: str,
        frames: List[FrameSummary],
        source_dir: Optional[Path] = None,
    ):
        if source_dir is not None:
            expanded_source_dir = path.expanduser(source_dir)
            self._message = self._strip_source_dir(message, expanded_source_dir)
            self._frames = [
                FrameSummary(
                    self._strip_source_dir(frame.filename, expanded_source_dir),
                    frame.lineno,
                    frame.name,
                    line=frame.line,
                )
                for frame in frames
            ]
        else:
            self._message = message
            self._frames = frames
        self._source_dir = source_dir

    @property
    def message(self) -> str:
        return self._message

    @property
    def frames(self) -> List[FrameSummary]:
        return self._frames

    @property
    def cause(self) -> str:
        return self._format_frame(self._frames[-1])

    @staticmethod
    def _format_frame(frame: FrameSummary) -> str:
        return f'File "{frame.filename}" on line {frame.lineno}: "{frame.line}"'

    @staticmethod
    def from_exception(
        exc: Exception, source_dir: Optional[Path] = None
    ) -> "PythonExecutionStatusReport":
        return PythonExecutionStatusReport(
            message=str(exc),
            frames=traceback.extract_tb(exc.__traceback__),
            source_dir=source_dir,
        )

    @staticmethod
    def _strip_source_dir(string: str, source_dir: str) -> str:
        return string.replace(source_dir + path.sep, "")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PythonExecutionStatusReport):
            return False
        return self.message == other.message and self.frames == other.frames


class AssertionFailureStatusReport(ExecutionStatusReport):
    def __init__(
        self,
        failed_assertions: Optional[List[Assertion]] = None,
        message: Optional[str] = None,
    ):
        if failed_assertions:
            stringified = ", ".join([str(assertion) for assertion in failed_assertions])
            self._message = f"failed: {stringified}"
        else:
            assert message
            self._message = message

    @property
    def message(self) -> str:
        return self._message

    @property
    def cause(self) -> str:
        return ""

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, AssertionFailureStatusReport):
            return False
        return self._message == other._message


class ExecutionStatusReportFactory:
    @staticmethod
    def from_json(json_str: str) -> ExecutionStatusReport:
        try:
            payload = json.loads(json_str)
        except json.JSONDecodeError:
            return GenericExecutionStatusReport(json_str)
        report_type = payload["type"].lower()
        if report_type == "python":
            message: str = payload["message"]
            frames: List[FrameSummary] = [
                FrameSummary(
                    filename=f["file"],
                    lineno=f["lineno"],
                    name=f["name"],
                    line=f["line"],
                )
                for f in payload["frames"]
            ]
            return PythonExecutionStatusReport(
                message=message,
                frames=frames,
            )
        elif report_type == "generic":
            return GenericExecutionStatusReport(payload["message"])
        elif report_type == "assertion":
            return AssertionFailureStatusReport(message=payload["message"])
        else:
            raise Exception("Invalid status report type")

    @staticmethod
    def to_json(report: ExecutionStatusReport) -> str:
        if isinstance(report, PythonExecutionStatusReport):
            dic_py = {
                "type": "Python",
                "message": report.message,
                "frames": [
                    {
                        "file": f.filename,
                        "line": f.line,
                        "lineno": f.lineno,
                        "name": f.name,
                    }
                    for f in report.frames
                ],
            }
            return json.dumps(dic_py)
        elif isinstance(report, GenericExecutionStatusReport):
            dic_generic = {"type": "Generic", "message": report.message}
            return json.dumps(dic_generic)
        elif isinstance(report, AssertionFailureStatusReport):
            dic_generic = {"type": "Assertion", "message": report.message}
            return json.dumps(dic_generic)
        else:
            raise Exception("Unsupported report type")



"""
This module is here to facilitate the transition to the new module name without breaking compatibility between the SDK and the backend
"""
# pylint: disable=unused-import
from .asset import AssetPath, AssetType, BaseAsset  # noqa



import os
import uuid
from typing import Any


_ENV_KEY_LAYER_DEBUG = "LAYER_DEBUG"
_ENV_KEY_REQUEST_ID = "REQUEST_ID"

_TRUTHY_LAYER_DEBUG_VALUES = [
    "True",
    "true",
    "1",
    1,
]


def is_layer_debug_on() -> bool:
    if _ENV_KEY_LAYER_DEBUG not in os.environ:
        return False

    return os.environ[_ENV_KEY_LAYER_DEBUG] in _TRUTHY_LAYER_DEBUG_VALUES


class UserSessionId:
    """
    UserSessionId is a singleton (only one instance per process,
    shared across threads) used to link all user requests by
    the same id.

    It will be used in grpc headers for remote interactions and
    propagated to downstream backend services to serve different
    purposes:
    - debugging a user journey: search DataDog logs by this id
    - reliably rely on retry mechanisms via idempotent requests
    """

    _instance = None

    def __new__(cls) -> Any:
        if cls._instance is None:
            id_ = uuid.uuid4()
            cls._instance = super(UserSessionId, cls).__new__(cls)
            cls._instance._value = id_  # type: ignore
        return cls._instance

    def __str__(self) -> str:
        return str(self._value)  # type: ignore



from types import TracebackType
from typing import Optional

from layer.contracts.assets import AssetType
from layer.contracts.datasets import DatasetBuild
from layer.tracker.progress_tracker import RunProgressTracker
from layer.training.base_train import BaseTrain


class Context:
    """
    Provides access to variables within the pipeline execution.

    This class should not be initialized by end-users.
    """

    def __init__(
        self,
        train: Optional[BaseTrain] = None,
        dataset_build: Optional[DatasetBuild] = None,
        tracker: Optional[RunProgressTracker] = None,
        asset_name: Optional[str] = None,
        asset_type: Optional[AssetType] = None,
    ) -> None:
        self._train: Optional[BaseTrain] = train
        self._dataset_build: Optional[DatasetBuild] = dataset_build
        self._tracker: Optional[RunProgressTracker] = tracker
        self._asset_name = asset_name
        self._asset_type = asset_type

    def train(self) -> Optional[BaseTrain]:
        """
        Retrieves the active Layer train object.

        :return: Represents the current train of the model, passed by Layer when the training of the model starts.

        .. code-block:: python

            # Get train object from Layer Context.
            train = context.train()
        """
        return self._train

    def dataset_build(self) -> Optional[DatasetBuild]:
        """
        Retrieves the active Layer dataset build object.

        :return: Represents the current dataset build of the dataset, passed by Layer when building of the dataset starts.
        """
        return self._dataset_build

    def with_train(self, train: Optional[BaseTrain]) -> None:
        self._train = train

    def with_dataset_build(self, dataset_build: Optional[DatasetBuild]) -> None:
        self._dataset_build = dataset_build

    def with_tracker(self, tracker: RunProgressTracker) -> None:
        self._tracker = tracker

    def with_asset_name(self, asset_name: str) -> None:
        self._asset_name = asset_name

    def with_asset_type(self, asset_type: AssetType) -> None:
        self._asset_type = asset_type

    def tracker(self) -> Optional[RunProgressTracker]:
        return self._tracker

    def asset_name(self) -> Optional[str]:
        return self._asset_name

    def asset_type(self) -> AssetType:
        if self._asset_type:
            return self._asset_type
        elif self.train():
            return AssetType.MODEL
        elif self.dataset_build():
            return AssetType.DATASET
        else:
            raise Exception("Unsupported asset type")

    def close(self) -> None:
        pass

    def __enter__(self) -> "Context":
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()



from pathlib import Path

import numpy as np
import pandas as pd
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor
from scipy.sparse import spmatrix  # type: ignore

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


class LightGBMModelFlavor(ModelFlavor):
    """An ML Model flavor implementation which handles persistence of LightGBM Models.
    Uses LightGBM model (an instance of lightgbm.Booster).

    """

    MODULE_KEYWORD = "lightgbm"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_LIGHTGBM

    def save_model_to_directory(
        self,
        model_object: ModelObject,
        directory: Path,
    ) -> None:
        import mlflow.lightgbm

        return mlflow.lightgbm.save_model(model_object, path=directory.as_posix())

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        import mlflow.lightgbm

        model = mlflow.lightgbm.load_model(directory.as_uri())
        return ModelRuntimeObjects(
            model, lambda input_df: self.__predict(model, input_df)
        )

    @staticmethod
    def __predict(model: ModelObject, input_df: pd.DataFrame) -> pd.DataFrame:
        prediction = model.predict(input_df)  # type: ignore
        if isinstance(prediction, np.ndarray):
            return pd.DataFrame(prediction)
        elif isinstance(prediction, pd.DataFrame):
            return prediction
        elif isinstance(prediction, spmatrix):
            return pd.DataFrame.sparse.from_spmatrix(prediction)  # type: ignore
        else:
            raise Exception(f"Unsupported return type: {type(prediction)}")



import inspect
import io
from logging import Logger
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import UUID

import pandas as pd
import requests  # type: ignore
from layerapi.api.value.logged_data_type_pb2 import LoggedDataType

from layer.clients.layer import LayerClient
from layer.contracts.logged_data import Markdown, ModelMetricPoint


if TYPE_CHECKING:
    import matplotlib.figure  # type: ignore
    import PIL.Image


class LogDataRunner:
    def __init__(
        self,
        client: LayerClient,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
        logger: Optional[Logger] = None,
    ):
        assert bool(train_id) ^ bool(dataset_build_id)
        self._client = client
        self._train_id = train_id
        self._dataset_build_id = dataset_build_id
        self._logger = logger

    def log(
        self,
        data: Dict[
            str,
            Union[
                str,
                float,
                bool,
                int,
                Dict[str, Any],
                pd.DataFrame,
                "PIL.Image.Image",
                "matplotlib.figure.Figure",
                ModuleType,
                Path,
            ],
        ],
        epoch: Optional[int] = None,
    ) -> None:
        LogDataRunner._check_epoch(epoch)

        for tag, value in data.items():
            if isinstance(value, str):
                self._log_text(tag=tag, text=value)
            # boolean check must be done before numeric check as it also returns true for booleans.
            elif isinstance(value, bool):
                self._log_boolean(tag=tag, bool_val=value)
            elif isinstance(value, (int, float)):
                if self._train_id and epoch is not None:
                    self._log_metric(tag=tag, numeric_value=value, epoch=epoch)
                else:
                    self._log_number(tag=tag, number=value)
            elif isinstance(value, Markdown):
                self._log_markdown(tag=tag, text=value.data)
            elif isinstance(value, pd.DataFrame):
                self._log_dataframe(tag=tag, df=value)
            elif LogDataRunner._is_tabular_dict(value):
                if TYPE_CHECKING:
                    assert isinstance(value, dict)
                dataframe = LogDataRunner._convert_dict_to_dataframe(value)
                self._log_dataframe(tag=tag, df=dataframe)
            elif LogDataRunner._is_video(value):
                if TYPE_CHECKING:
                    assert isinstance(value, Path)
                self._log_video_from_path(tag=tag, path=value)
            elif LogDataRunner._is_image(value):
                if TYPE_CHECKING:
                    assert isinstance(value, Path)
                if self._train_id and epoch is not None:
                    self._log_image_from_path(tag=tag, path=value, epoch=epoch)
                else:
                    self._log_image_from_path(tag=tag, path=value)
            elif self._is_pil_image(value):
                if TYPE_CHECKING:
                    import PIL.Image

                    assert isinstance(value, PIL.Image.Image)
                if self._train_id and epoch is not None:
                    self._log_image(tag=tag, image=value, epoch=epoch)
                else:
                    self._log_image(tag=tag, image=value)
            elif self._is_plot_figure(value):
                if TYPE_CHECKING:
                    import matplotlib.figure

                    assert isinstance(value, matplotlib.figure.Figure)
                self._log_plot_figure(tag=tag, figure=value)
            elif self._is_axes_subplot(value):
                if TYPE_CHECKING:
                    import matplotlib.axes._subplots  # type: ignore

                    assert isinstance(
                        value,
                        matplotlib.axes._subplots.AxesSubplot,  # pylint: disable=protected-access
                    )
                self._log_plot_figure(tag=tag, figure=value.get_figure())
            elif self._is_pyplot(value):
                assert isinstance(value, ModuleType)
                self._log_current_plot_figure(tag=tag, plt=value)
            else:
                raise ValueError(f"Unsupported value type -> {type(value)}")

    def _log_metric(
        self, tag: str, numeric_value: Union[float, int], epoch: Optional[int]
    ) -> None:
        assert self._train_id
        # store numeric values w/o an explicit epoch as metric with the special epoch:-1
        epoch = epoch if epoch is not None else -1
        self._client.logged_data_service_client.log_model_metric(
            train_id=self._train_id,
            tag=tag,
            points=[ModelMetricPoint(epoch=epoch, value=float(numeric_value))],
        )

    def _log_text(self, tag: str, text: str) -> None:
        self._client.logged_data_service_client.log_text_data(
            train_id=self._train_id,
            dataset_build_id=self._dataset_build_id,
            tag=tag,
            data=text,
        )

    def _log_markdown(self, tag: str, text: str) -> None:
        self._client.logged_data_service_client.log_markdown_data(
            train_id=self._train_id,
            dataset_build_id=self._dataset_build_id,
            tag=tag,
            data=text,
        )

    def _log_number(self, tag: str, number: Union[float, int]) -> None:
        self._client.logged_data_service_client.log_numeric_data(
            train_id=self._train_id,
            dataset_build_id=self._dataset_build_id,
            tag=tag,
            data=str(number),
        )

    def _log_boolean(self, tag: str, bool_val: bool) -> None:
        self._client.logged_data_service_client.log_boolean_data(
            train_id=self._train_id,
            dataset_build_id=self._dataset_build_id,
            tag=tag,
            data=str(bool_val),
        )

    def _log_dataframe(self, tag: str, df: pd.DataFrame) -> None:
        rows = len(df.index)
        if rows > 1000:
            raise ValueError(
                f"DataFrame rows size cannot exceed 1000. Current size: {rows}"
            )
        df_json = df.to_json(orient="table")  # type: ignore
        self._client.logged_data_service_client.log_table_data(
            train_id=self._train_id,
            dataset_build_id=self._dataset_build_id,
            tag=tag,
            data=df_json,
        )

    def _log_video_from_path(self, tag: str, path: Path) -> None:
        self._log_binary_from_path(
            tag, path, LoggedDataType.LOGGED_DATA_TYPE_VIDEO, max_file_size_mb=100
        )

    def _log_image_from_path(
        self, tag: str, path: Path, epoch: Optional[int] = None
    ) -> None:
        self._log_binary_from_path(
            tag,
            path,
            LoggedDataType.LOGGED_DATA_TYPE_IMAGE,
            max_file_size_mb=1,
            epoch=epoch,
        )

    def _log_binary_from_path(
        self,
        tag: str,
        path: Path,
        logged_data_type: "LoggedDataType.V",
        *,
        max_file_size_mb: int,
        epoch: Optional[int] = None,
    ) -> None:
        file_size_in_bytes = path.stat().st_size
        self._check_size_less_than_mb(file_size_in_bytes, max_file_size_mb)
        with requests.Session() as s, open(path, "rb") as image_file:
            presigned_url = self._client.logged_data_service_client.log_binary_data(
                train_id=self._train_id,
                dataset_build_id=self._dataset_build_id,
                tag=tag,
                logged_data_type=logged_data_type,
                epoch=epoch,
            )
            resp = s.put(presigned_url, data=image_file)
            resp.raise_for_status()

    def _log_image(
        self, tag: str, image: "PIL.Image.Image", *, epoch: Optional[int] = None
    ) -> None:
        with requests.Session() as s, io.BytesIO() as buffer:
            if image.mode in ["RGBA", "P"]:
                image.save(buffer, format="PNG")
            else:
                image.save(buffer, format="JPEG")
            self._check_buffer_size(buffer=buffer)
            presigned_url = self._client.logged_data_service_client.log_binary_data(
                train_id=self._train_id,
                dataset_build_id=self._dataset_build_id,
                tag=tag,
                logged_data_type=LoggedDataType.LOGGED_DATA_TYPE_IMAGE,
                epoch=epoch,
            )
            resp = s.put(presigned_url, data=buffer.getvalue())
            resp.raise_for_status()

    def _log_plot_figure(self, tag: str, figure: "matplotlib.figure.Figure") -> None:
        with requests.Session() as s, io.BytesIO() as buffer:
            figure.savefig(buffer, format="jpg")
            self._check_buffer_size(buffer=buffer)
            presigned_url = self._client.logged_data_service_client.log_binary_data(
                train_id=self._train_id,
                dataset_build_id=self._dataset_build_id,
                tag=tag,
                logged_data_type=LoggedDataType.LOGGED_DATA_TYPE_IMAGE,
            )
            resp = s.put(presigned_url, data=buffer.getvalue())
            resp.raise_for_status()

    def _log_current_plot_figure(self, tag: str, plt: ModuleType) -> None:
        if len(plt.get_fignums()) > 0:
            self._log_plot_figure(tag, plt.gcf())
        else:
            raise ValueError("No figures in the current pyplot state!")

    @staticmethod
    def _check_epoch(epoch: Any) -> None:
        if epoch is not None:
            if not isinstance(epoch, int) or epoch < 0:
                raise ValueError(
                    f"epoch (i.e. step) can only be a non-negative integer, given value: {epoch}"
                )

    @staticmethod
    def _has_allowed_extension(
        file: Path, allowed_extensions: Optional[List[str]]
    ) -> bool:
        if allowed_extensions is None:
            allowed_extensions = []
        extension = file.suffix.lower()
        for allowed_extension in allowed_extensions:
            if extension == allowed_extension:
                return True
        return False

    @staticmethod
    def _convert_dict_to_dataframe(dictionary: Dict[str, Any]) -> pd.DataFrame:
        new_values = []
        for value in dictionary.values():
            if isinstance(value, (float, int, str, bool)):
                new_values.append(value)
            else:
                new_values.append(str(value))
        df = pd.DataFrame({"name": dictionary.keys(), "value": new_values})  # type: ignore
        df = df.set_index("name")
        return df

    @staticmethod
    def _is_tabular_dict(maybe_dict: Any) -> bool:
        if not isinstance(maybe_dict, dict):
            return False

        for key in maybe_dict:
            if not isinstance(key, str):
                return False

        return True

    @staticmethod
    def _is_image(value: Any) -> bool:
        return isinstance(value, Path) and LogDataRunner._has_allowed_extension(
            value, [".gif", ".png", ".jpg", ".jpeg"]
        )

    @staticmethod
    def _is_video(value: Any) -> bool:
        return isinstance(value, Path) and LogDataRunner._has_allowed_extension(
            value, [".mp4", ".webm", ".ogg"]
        )

    def _is_pil_image(self, value: Any) -> bool:
        return "PIL.Image" in self._get_base_module_list(value)

    def _is_plot_figure(self, value: Any) -> bool:
        return "matplotlib.figure" in self._get_base_module_list(value)

    def _is_axes_subplot(self, value: Any) -> bool:
        base_module_list = self._get_base_module_list(value)
        return "matplotlib.axes._subplots" in base_module_list

    def _is_pyplot(self, value: Any) -> bool:
        return (
            hasattr(value, "__name__")
            and "matplotlib.pyplot" == value.__name__
            and isinstance(value, ModuleType)
        )

    def _get_base_module_list(self, value: Any) -> List[str]:
        return [
            inspect.getmodule(clazz).__name__ for clazz in inspect.getmro(type(value))  # type: ignore
        ]

    def _check_buffer_size(self, buffer: io.BytesIO) -> None:
        size_in_bytes = buffer.tell()
        self._check_size_less_than_mb(size_in_bytes, 1)

    def _check_size_less_than_mb(
        self, size_in_bytes: float, max_mb_size: float
    ) -> None:
        size_in_mb = size_in_bytes / 1000**2
        if size_in_mb > max_mb_size:
            raise ValueError(
                f"Image size cannot exceed {max_mb_size}MB. Current size: {size_in_mb}"
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['layer',
 'layer.cache',
 'layer.clients',
 'layer.config',
 'layer.contracts',
 'layer.decorators',
 'layer.exceptions',
 'layer.flavors',
 'layer.logged_data',
 'layer.projects',
 'layer.tracker',
 'layer.training',
 'layer.training.runtime',
 'layer.utils',
 'layer.utils.grpc']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.3',
 'aiohttp>=3.7.3,<3.8.0',
 'boto3>=1.16.24',
 'cloudpickle>=2.0.0',
 'cryptography>=3.4.7',
 'docker>=4,<5',
 'grpcio-tools==1.45.0',
 'grpcio==1.45.0',
 'humanize>=3.11.0',
 'idna<3',
 'jsonschema==3.1.1',
 'layer-api==0.9.366764',
 'luddite==1.0.2',
 'mlflow>=1.25.0',
 'networkx>=2.5',
 'packaging<=21.0',
 'pandas==1.3.5',
 'polling>=0.3.1',
 'protobuf>=3.12.0',
 'pyarrow==7.0.0',
 'pyjwt>=2.0.0,<3.0.0',
 'rich>=11',
 'transformers',
 'validate-email==1.3',
 'wrapt>=1.13.3',
 'yarl>=1.6.3']

extras_require = \
{':python_version < "3.8"': ['pickle5>=0.0.11,<0.1.0'],
 ':python_version >= "3.7" and python_version < "3.8"': ['numpy<1.22']}

setup_kwargs = {
    'name': 'layer',
    'version': '0.10.2548775939',
    'description': 'Layer AI SDK',
    'long_description': '<!---\nCopyright 2022 Layer. All rights reserved.\n\nLicensed under the Apache License, Version 2.0 (the "License");\nyou may not use this file except in compliance with the License.\nYou may obtain a copy of the License at\n\n    http://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software\ndistributed under the License is distributed on an "AS IS" BASIS,\nWITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\nSee the License for the specific language governing permissions and\nlimitations under the License.\n-->\n\n<p align="center">\n    <br>\n    <a href="https://layer.ai">\n          <img src="https://app.layer.ai/assets/layer_wordmark_black.png" width="40%"\nalt="Layer"/>\n    </a>\n    <br>\n<p>\n<p align="center">\n    <a href="https://github.com/layerai/sdk/blob/main/LICENSE">\n        <img alt="License" src="https://img.shields.io/github/license/layerai/sdk.svg?color=blue">\n    </a>\n    <a href="https://docs.app.layer.ai">\n        <img alt="Documentation" src="https://img.shields.io/badge/docs-online-success">\n    </a>\n    <a href="https://github.com/layerai/sdk/actions/workflows/release.yml">\n        <img alt="Build" src="https://img.shields.io/github/workflow/status/layerai/sdk/Release">\n    </a>\n    <a href="https://pypi.python.org/pypi/layer">\n        <img alt="PyPI" src="https://img.shields.io/pypi/v/layer.svg">\n    </a>\n    <a href="https://github.com/layerai/.github/blob/main/CODE_OF_CONDUCT.md">\n        <img alt="Contributor Covenant" src="https://img.shields.io/badge/contributor%20covenant-v2.1%20adopted-blueviolet.svg">\n    </a>\n</p>\n\n## Layer - Metadata Store for Production ML\n\n![Layer - Metadata store for production ML](https://app.layer.ai/assets/layer_metadata_store.png)\n\n\n[Layer](https://layer.ai) helps you build, train and track all your machine learning project metadata including ML models and datasets with semantic versioning, extensive artifact logging and dynamic reporting with local↔cloud training\n\n**[Start for Free now!](https://app.layer.ai/login?returnTo=%2Fgetting-started)**\n\n## Getting Started\n\nInstall Layer:\n```shell\npip install layer --upgrade\n```\n\nLogin to your free account and initialize your project:\n```python\nimport layer\nlayer.login()\nlayer.init("my-first-project")\n```\n\nDecorate your training function to register your model to Layer:\n```python\nfrom layer.decorators import model\n\n@model("my-model")\ndef train():\n    from sklearn import datasets\n    from sklearn.svm import SVC\n    iris = datasets.load_iris()\n    clf = SVC()\n    clf.fit(iris.data, iris.target)\n    return clf\n\ntrain()\n```\n\nNow you can fetch your model from Layer:\n\n```python\nimport layer\n\nclf = layer.get_model("my-model:1.1").get_train()\nclf\n\n# > SVC()\n```\n\n[**🚀 Try in Google Colab now!**](https://colab.research.google.com/github/layerai/examples/blob/main/tutorials/add-models-to-layer/how_to_add_models_to_layer.ipynb)\n\n## Reporting bugs\nYou have a bug, a request or a feature? Let us know on [Slack](https://bit.ly/layercommunityslack) or [open an issue](https://github.com/layerai/sdk/issues/new/choose)\n\n## Contributing code\nDo you want to help us build the best metadata store? Check out the [Contributing Guide](https://github.com/layerai/sdk/blob/main/CONTRIBUTING.md)\n\n## Learn more\n- Join our [Slack Community ](https://bit.ly/layercommunityslack) to connect with other Layer users\n- Visit the [examples repo](https://github.com/layerai/examples) for more inspiration\n- Browse [Community Projects](https://layer.ai/community) to see more use cases\n- Check out the [Documentation](https://docs.app.layer.ai)\n- [Contact us](https://layer.ai/contact-us) for your questions\n',
    'author': 'Layer',
    'author_email': 'info@layer.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://layer.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



from typing import TYPE_CHECKING, Any
from uuid import UUID

from layer.tracker.progress_tracker import RunProgressTracker


if TYPE_CHECKING:
    from layer.types import ModelObject


class BaseTrain:
    """
    Allows the user to control the state of a model training.

    This class should never be instantiated directly by end users.

    .. code-block:: python

        def train_model(train: Train, tf: Featureset("transaction_features")):
            # We create the training and label data
            train_df = tf.to_pandas()
            X = train_df.drop(["transactionId", "is_fraud"], axis=1)
            Y = train_df["is_fraud"]
            # user model training will happen below [...]
    """

    def get_id(self) -> UUID:
        pass

    def get_version(self) -> str:
        pass

    def get_train_index(self) -> str:
        pass

    def save_model(
        self,
        trained_model_obj: "ModelObject",
        tracker: RunProgressTracker,
    ) -> Any:
        pass

    def __start_train(self) -> None:
        pass

    def __complete_train(self) -> None:
        pass

    def __enter__(self) -> Any:
        pass

    def __exit__(
        self, exception_type: Any, exception_value: Any, traceback: Any
    ) -> None:
        pass
